VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "UniSock"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
'*************************************************************************************************
'* UniSock - Winsock replacement
'* -----------------------------
'* By Vesa Piittinen aka Merri, http://vesa.piittinen.name/ <vesa@piittinen.name>
'*
'* LICENSE
'* -------
'* http://creativecommons.org/licenses/by-sa/1.0/fi/deed.en
'*
'* Terms: 1) If you make your own version, share using this same license.
'*        2) When used in a program, mention my name in the program's credits.
'*        3) Free for commercial and non-commercial usage.
'*        4) Use at your own risk. No support guaranteed.
'*
'* REQUIREMENTS
'* ------------
'* Huh what? Just this one class module. No extra files required.
'*
'* VERSION HISTORY
'* ---------------
'* Version 1.2 (2008-08-12)
'* - Vista connection error resolved!
'*
'* Version 1.1 (2008-08-07)
'* - CreateWindowExW fixed: had forgotten to call DefWindowProc, no more subclassing!
'* - Fixed crash bug when using an invalid RequestID value for Accept.
'* - No more blank error messages, all errors should be covered.
'* - Unicode support: you may now send and receive Unicode! You can get into this mode by using Mode = [Socket Text Mode]
'*   - event: TextArrival
'*   - property: Mode (default: [Socket Binary Mode])
'*   - method: SendText (default: UTF-8)
'*   How it works? Well, first of all: you do not need to care about line changes. SendText adds line change to each line.
'*   TextArrival also processes data one line at a time. Unicode detection is done by looking for line changes.
'*   For example, [CR][00][LF][00] guarantees the data is UTF-16. [CR][LF] guarantees data is either UTF-8 or ANSI.
'*   If data is UTF-16 or UTF-8 you get it in "ready to use" format. ANSI however is provided as raw data (use StrConv).
'*
'* Version 1.0 (2008-08-06)
'* - All the Winsock control features implemented.
'* - Changes & extra features:
'*   - Autocomplete: datatypes have changed in some events and methods to be more precise.
'*   - GetData and SendData support a wide range of array datatypes. (Including string arrays!)
'*   - LineChange: line change used when sending string arrays.
'*   - RequestHost, RequestIP and RequestPort must be used instead of Remote ones when using Accept event.
'*   - Some methods return True on success.
'*
'* Important: UDP SUPPORT IS UNTESTED. Bug reports with sample projects welcome!
'*
'* CREDITS
'* -------
'* Paul Caton and LaVolpe for their work on SelfSub, SelfHook and SelfCallback
'*************************************************************************************************
Option Explicit

Public Event Closing()
Public Event Connect()
Public Event ConnectionRequest(ByVal RequestID As Long)
Public Event DataArrival(ByVal BytesTotal As Long)
Public Event Error(ByVal Number As Integer, Description As String, ByVal sCode As Long, ByVal Source As String, ByVal HelpFile As String, ByVal HelpContext As Long, CancelDisplay As Boolean)
Public Event SendComplete()
Public Event SendProgress(ByVal BytesSent As Long, ByVal BytesRemaining As Long)
Public Event StatusChange(ByVal Status As StateConstants)
Public Event TextArrival(Text As String, ByVal LineChange As UniSockLineChange, ByVal ANSI As Boolean)

' for compatibility...
Public Enum ErrorConstants
    sckAddressInUse = 10048
    sckAddressNotAvailable = 10049
    sckAlreadyComplete = 10037
    sckAlreadyConnected = 10056
    sckBadState = 40006
    sckConnectAborted = 10053
    sckConnectionRefused = 10061
    sckConnectionReset = 10054
    sckGetNotSupported = 394
    sckHostNotFound = 11001
    sckHostNotFoundTryAgain = 11002
    sckInProgress = 10036
    sckInvalidArg = 40014
    sckInvalidArgument = 10014
    sckInvalidOp = 40020
    sckInvalidPropertyValue = 380
    sckMsgTooBig = 10040
    sckNetReset = 10052
    sckNetworkSubsystemFailed = 10050
    sckNetworkUnreachable = 10051
    sckNoBufferSpace = 10055
    sckNoData = 11004
    sckNonRecoverableError = 11003
    sckNotConnected = 10057
    sckNotInitialized = 10093
    sckNotSocket = 10038
    sckOpCanceled = 10004
    sckOutOfMemory = 7
    sckOutOfRange = 40021
    sckPortNotSupported = 10043
    sckSetNotSupported = 383
    sckSocketShutdown = 10058
    sckSuccess = 40017
    sckTimedout = 10060
    sckUnsupported = 40018
    sckWouldBlock = 10035
    sckWrongProtocol = 40026
End Enum

' for compatibility...
Public Enum ProtocolConstants
    sckTCPProtocol
    sckUDPProtocol
End Enum

' for compatibility...
Public Enum StateConstants
    sckClosed
    sckOpen
    sckListening
    sckConnectionPending
    sckResolvingHost
    sckHostResolved
    sckConnecting
    sckConnected
    sckClosing
    sckError
End Enum

Public Enum UniSockLineChange
    [Socket Null]
    [Socket CRLF]
    [Socket LF]
    [Socket CR]
End Enum

Public Enum UniSockMode
    [Socket Binary Mode]
    [Socket Text Mode]
End Enum

Public Enum UniSockTextFormat
    [Socket UTF-8]
    [Socket ANSI]
    [Socket UTF-16]
End Enum

Private Type HOSTENT
    hName As Long
    hAliases As Long
    hAddrType As Integer
    hLength As Integer
    hAddrList As Long
End Type

Private Type SOCKADDR_IN
    Family As Integer
    Port As Integer
    Addr As Long
    Zero(7) As Byte
End Type

Private Type WNDCLASS
    Style As Long
    lpfnWndProc As Long
    cbClsextra As Long
    cbWndExtra2 As Long
    hInstance As Long
    hIcon As Long
    hCursor As Long
    hbrBackground As Long
    lpszMenuName As Long
    lpszClassName As Long
End Type

Private Type WSAData
    wVersion As Integer
    wHighVersion As Integer
    szDescription(256) As Byte
    szSystemStatus(128) As Byte
    iMaxSockets As Integer
    iMaxUdpDg As Integer
    lpVendorInfo As Long
End Type

Private Declare Function CreateWindowExW Lib "user32" (ByVal dwExStyle As Long, ByVal lpClassName As Long, ByVal lpWindowName As Long, ByVal dwStyle As Long, ByVal X As Long, ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal hWndParent As Long, ByVal hMenu As Long, ByVal hInstance As Long, lpParam As Any) As Long
Private Declare Function DefWindowProcW Lib "user32" (ByVal hWnd As Long, ByVal wMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
Private Declare Function DestroyWindow Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function FindWindowW Lib "user32" (ByVal lpClassName As Long, ByVal lpWindowName As Long) As Long
Private Declare Function FormatMessageW Lib "kernel32" (ByVal dwFlags As Long, lpSource As Any, ByVal dwMessageId As Long, ByVal dwLanguageId As Long, ByVal lpBuffer As Long, ByVal nSize As Long, Arguments As Long) As Long
Private Declare Function GetLastError Lib "kernel32" () As Long
Private Declare Function GlobalAlloc Lib "kernel32" (ByVal wFlags As Long, ByVal dwBytes As Long) As Long
Private Declare Function GlobalFree Lib "kernel32" (ByVal hMem As Long) As Long
Private Declare Function GlobalLock Lib "kernel32" (ByVal hMem As Long) As Long
Private Declare Function GlobalUnlock Lib "kernel32" (ByVal hMem As Long) As Long
Private Declare Function PostMessageW Lib "user32" (ByVal hWnd As Long, ByVal wMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
Private Declare Function RegisterClassW Lib "user32" (Class As WNDCLASS) As Long
Private Declare Function SetWindowLongW Lib "user32" (ByVal hWnd As Long, ByVal nIndex As Long, ByVal dwNewLong As Long) As Long
Private Declare Function UnregisterClassW Lib "user32" (ByVal lpClassName As Long, ByVal hInstance As Long) As Long

Private Declare Function lstrlenA Lib "kernel32" (ByVal lpString As Long) As Long
Private Declare Function MultiByteToWideChar Lib "kernel32" (ByVal CodePage As Long, ByVal dwFlags As Long, ByVal lpMultiByteStr As Long, ByVal cchMultiByte As Long, ByVal lpWideCharStr As Long, ByVal cchWideChar As Long) As Long
Private Declare Sub PutMem4 Lib "msvbvm60" (Destination As Any, Value As Any)
Private Declare Sub PutMem8 Lib "msvbvm60" (Destination As Any, Value As Any)
Private Declare Sub RtlMoveMemory Lib "kernel32" (Destination As Any, Source As Any, ByVal Length As Long)
Private Declare Function SysAllocStringANSI Lib "oleaut32" Alias "SysAllocStringByteLen" (ByVal OleStr As Long, ByVal bLen As Long) As String
Private Declare Function SysAllocStringByteLen Lib "oleaut32" (ByVal OleStr As Long, ByVal bLen As Long) As Long
Private Declare Function SysAllocStringLen Lib "oleaut32" (ByVal OleStr As Long, ByVal bLen As Long) As Long
Private Declare Function WideCharToMultiByte Lib "kernel32" (ByVal CodePage As Long, ByVal dwFlags As Long, ByVal lpWideCharStr As Long, ByVal cchWideChar As Long, ByVal lpMultiByteStr As Long, ByVal cchMultiByte As Long, ByVal lpDefaultChar As Long, lpUsedDefaultChar As Long) As Long

Private Declare Function SocketAccept Lib "ws2_32" Alias "accept" (ByVal S As Long, ByRef Addr As SOCKADDR_IN, ByRef AddrLen As Long) As Long
Private Declare Function SocketBind Lib "ws2_32" Alias "bind" (ByVal S As Long, ByRef Name As SOCKADDR_IN, ByRef NameLen As Long) As Long
Private Declare Function SocketClose Lib "ws2_32" Alias "closesocket" (ByVal S As Long) As Long
Private Declare Function SocketConnect Lib "ws2_32" Alias "connect" (ByVal S As Long, ByRef Name As SOCKADDR_IN, ByVal NameLen As Long) As Long
Private Declare Function SocketGetHostName Lib "ws2_32" Alias "gethostname" (ByVal Hostname As String, ByVal NameLen As Long) As Long
Private Declare Function SocketGetHostByName Lib "ws2_32" Alias "gethostbyname" (ByVal Hostname As String) As Long
Private Declare Function SocketGetIP Lib "ws2_32" Alias "inet_addr" (ByVal CP As String) As Long
Private Declare Function SocketGetName Lib "ws2_32" Alias "getsockname" (ByVal S As Long, ByRef Name As SOCKADDR_IN, ByRef NameLen As Long) As Long
Private Declare Function SocketGetNameInfo Lib "ws2_32" Alias "getnameinfo" (ByRef SA As SOCKADDR_IN, ByVal SALen As Long, ByVal Host As String, ByVal HostLen As Long, ByVal Serv As String, ByVal ServLen As Long, ByVal Flags As Long) As Long
Private Declare Function SocketGetOptions Lib "ws2_32" Alias "getsockopt" (ByVal S As Long, ByVal Level As Long, ByVal OptName As Long, OptVal As Any, OptLen As Long) As Long
Private Declare Function SocketGetPeerName Lib "ws2_32" Alias "getpeername" (ByVal S As Long, ByRef Name As SOCKADDR_IN, ByRef NameLen As Long) As Long
Private Declare Function SockethTons Lib "ws2_32" Alias "htons" (ByVal hostshort As Integer) As Integer
Private Declare Function SocketInet_nToa Lib "ws2_32" Alias "inet_ntoa" (ByVal in_addr As Long) As Long
Private Declare Function SocketInit Lib "ws2_32" Alias "socket" (ByVal af As Long, ByVal s_type As Long, ByVal Protocol As Long) As Long
Private Declare Function SocketIOCtl Lib "ws2_32" Alias "ioctlsocket" (ByVal S As Long, ByVal Cmd As Long, ByRef ArgP As Long) As Long
Private Declare Function SocketListen Lib "ws2_32" Alias "listen" (ByVal S As Long, ByVal BackLog As Long) As Long
Private Declare Function SocketnTons Lib "ws2_32" Alias "htons" (ByVal netshort As Integer) As Integer
Private Declare Function SocketReceive Lib "ws2_32" Alias "recv" (ByVal S As Long, Buf As Any, ByVal BufLen As Long, ByVal Flags As Long) As Long
Private Declare Function SocketReceiveFrom Lib "ws2_32" Alias "recv" (ByVal S As Long, Buf As Any, ByVal BufLen As Long, ByVal Flags As Long, ByRef FromAddr As SOCKADDR_IN, ByVal FromLen As Long) As Long
Private Declare Function SocketSend Lib "ws2_32" Alias "send" (ByVal S As Long, Buf As Any, ByVal BufLen As Long, ByVal Flags As Long) As Long
Private Declare Function SocketSendTo Lib "ws2_32" Alias "sendto" (ByVal S As Long, Buf As Any, ByVal BufLen As Long, ByVal Flags As Long, ByRef ToAddr As SOCKADDR_IN, ByVal ToLen As Long) As Long
Private Declare Function SocketSetOptions Lib "ws2_32" Alias "setsockopt" (ByVal S As Long, ByVal Level As Long, ByVal OptName As Long, OptVal As Any, ByVal OptLen As Long) As Long

Private Declare Function WSAAsyncGetHostByName Lib "ws2_32" (ByVal hWnd As Long, ByVal wMsg As Long, ByVal strHostName As String, Buf As Any, ByVal BufLen As Long) As Long
Private Declare Function WSAAsyncSelect Lib "ws2_32" (ByVal S As Long, ByVal hWnd As Long, ByVal wMsg As Long, ByVal lEvent As Long) As Long
Private Declare Function WSACancelAsyncRequest Lib "ws2_32" (ByVal hAsyncTaskHandle As Long) As Long
Private Declare Function WSACleanup Lib "ws2_32" () As Long
Private Declare Function WSAGetLastError Lib "ws2_32" () As Long
Private Declare Function WSAStartup Lib "ws2_32" (ByVal wVersionRequired As Long, lpWSADATA As WSAData) As Long

Private Const AF_INET = &H2&
Private Const FD_READ = &H1&
Private Const FD_WRITE = &H2&
Private Const FD_ACCEPT = &H8&
Private Const FD_CONNECT = &H10&
Private Const FD_CLOSE = &H20&
Private Const FIONREAD = &H4004667F
Private Const INADDR_NONE = -1&
Private Const INVALID_SOCKET = -1&
Private Const IPPROTO_TCP = 6&
Private Const IPPROTO_UDP = 17&
Private Const MSG_PEEK = &H2&
Private Const SO_BROADCAST = &H20&
Private Const SO_MAX_MSG_SIZE = &H2003&
Private Const SO_RCVBUF = &H1002&
Private Const SO_SNDBUF = &H1001&
Private Const SOCK_DGRAM = 2&
Private Const SOCK_STREAM = 1&
Private Const SOCKET_ERR = -1&
Private Const SOL_SOCKET = &HFFFF&
Private Const SOMAXCONN = 5&

Private Const WM_USER = &H400
Private Const RESOLVE_MESSAGE = WM_USER + &H400
Private Const SOCKET_MESSAGE = WM_USER + &H401

Private m_AsyncHost As Long
Private m_Buffer As String
Private m_BufferSize As Long
Private m_Class As WNDCLASS
Private m_ClassName As String
Private m_HostentHandle As Long
Private m_HostentPointer As Long
Private m_hWnd As Long
Private m_Initialized As Boolean
Private m_LineChange As String
Private m_LocalIP As String
Private m_LocalPort As Long
Private m_LocalPortBind As Long
Private m_Mode As UniSockMode
Private m_Peek As Boolean
Private m_Protocol As ProtocolConstants
Private m_ReceiveSize As Long
Private m_RegisterClass As Long
Private m_RemoteHost As String
Private m_RemoteIP As Long
Private m_RemotePort As Long
Private m_Request As Collection
Private m_RequestBuffer As String
Private m_RequestID As Long
Private m_RequestSA As SOCKADDR_IN
Private m_SendBuffer As String
Private m_SendBufferSize As Long
Private m_SendSize As Long
Private m_Socket As Long
Private m_State As StateConstants
Private m_WSAData As WSAData

Private CRLF As String
Private CR As String
Private LF As String
Private NULLCHAR As String

' the following are the requirements for SelfCallback
Private Declare Sub GetMem1 Lib "msvbvm60" (ByVal Addr As Long, RetVal As Byte)
Private Declare Sub GetMem4 Lib "msvbvm60" (ByVal Addr As Long, RetVal As Long)
Private Declare Function GetModuleHandleA Lib "kernel32" (ByVal lpModuleName As String) As Long
Private Declare Function GetProcAddress Lib "kernel32" (ByVal hModule As Long, ByVal lpProcName As String) As Long
Private Declare Function IsBadCodePtr Lib "kernel32" (ByVal lpfn As Long) As Long
Private Declare Sub PutMem2 Lib "msvbvm60" (ByVal Addr As Long, ByVal NewVal As Integer)
Private Declare Sub RtlMachineCodeCopy Lib "kernel32" Alias "RtlMoveMemory" (Destination As Long, Source As MachineCode, ByVal Length As Long)
Private Declare Function VirtualAlloc Lib "kernel32" (ByVal lpAddress As Long, ByVal dwSize As Long, ByVal flAllocationType As Long, ByVal flProtect As Long) As Long
Private Declare Function VirtualFree Lib "kernel32" (ByVal lpAddress As Long, ByVal dwSize As Long, ByVal dwFreeType As Long) As Long

' self-documentation: less comments and constants
Private Type MachineCode        ' 37 * 4 = 148 bytes
    ' thunk
    OwnerPtr As Long            ' 0
    CallbackAddress As Long     ' 1
    API_EbMode As Long          ' 2
    API_IsBadCodePtr As Long    ' 3
    API_KillTimer As Long       ' 4
    ' code
    MC1(5 To 5) As Long         ' 5
    AllocatedDataPtr As Long    ' 6
    MC2(7 To 17) As Long        ' 7 - 17
    ParamCount As Long          ' 18
    MC3(19 To 35) As Long       ' 19 - 35
    Ordinal As Long             ' 36
End Type

Public Function Accept(ByVal RequestID As Long) As Boolean
    Dim clsSocket As UniSock, clsTemp As UniSock, lngBuffer As Long, udtSockAddr As SOCKADDR_IN
    ' validate
    If (m_State = sckClosed) And (RequestID <> 0) Then
        ' ensure this RequestID is from an existing UniSock
        If FindWindowW(0, StrPtr(CStr(RequestID))) Then
            ' make sure socket is closed
            CloseSocket
            ' make a reference to the listening object
            PutMem4 clsTemp, ByVal RequestID
            Set clsSocket = clsTemp
            PutMem4 clsTemp, ByVal 0&
            ' now copy the data, get real RequestID etc.
            clsSocket.Friend_Request m_hWnd, m_RequestID, m_RequestSA, m_Buffer
            ' we do not need it for more than that
            Set clsSocket = Nothing
            ' correct some settings
            m_Socket = m_RequestID
            m_BufferSize = LenB(m_Buffer)
            m_Protocol = sckTCPProtocol
            If SocketGetOptions(m_RequestID, SOL_SOCKET, SO_RCVBUF, lngBuffer, 4) <> SOCKET_ERR Then m_ReceiveSize = lngBuffer
            If SocketGetOptions(m_RequestID, SOL_SOCKET, SO_SNDBUF, lngBuffer, 4) <> SOCKET_ERR Then m_SendSize = lngBuffer
            ' register messages
            If WSAAsyncSelect(m_RequestID, m_hWnd, SOCKET_MESSAGE, FD_READ Or FD_WRITE Or FD_ACCEPT Or FD_CONNECT Or FD_CLOSE) = SOCKET_ERR Then
                Private_SetState sckError, "Accept", WSAGetLastError
            End If
            ' local information
            If SocketGetName(m_RequestID, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR Then
                m_LocalIP = CStr(udtSockAddr.Addr And &HFF&) & "." & _
                    CStr((udtSockAddr.Addr And &HFF00&) \ &H100&) & "." & _
                    CStr((udtSockAddr.Addr And &HFF0000) \ &H10000) & "." & _
                    CStr(((udtSockAddr.Addr And &HFF000000) \ &H1000000) And &HFF&)
                m_LocalPortBind = SocketnTons(udtSockAddr.Port) And &HFFFF&
            Else
                Private_SetState sckError, "Accept", WSAGetLastError
            End If
            ' remote information
            m_RemoteHost = RequestHost
            m_RemoteIP = m_RequestSA.Addr
            m_RemotePort = SocketnTons(m_RequestSA.Port) And &HFFFF&
            ' done!
            m_RequestID = 0
            m_RequestSA = udtSockAddr
            ' status changes
            Private_SetState sckConnected
            RaiseEvent Connect
            ' last but not least, data has arrived
            If m_BufferSize Then Private_Message FD_READ
            ' success!
            Accept = True
        Else
            Err.Raise sckInvalidArg, "UniSock.Accept", Private_FormatWSAError(sckInvalidArg)
        End If
    Else
        Err.Raise sckInvalidOp, "UniSock.Accept", Private_FormatWSAError(sckInvalidOp)
    End If
End Function

Public Function Bind(Optional ByVal LocalPort As Long, Optional LocalIP As String) As Boolean
    Dim lngBuffer As Long
    If m_Initialized Then
        ' create socket if we do not have one
        If m_Socket = INVALID_SOCKET Then
            ' create socket for that protocol
            If m_Protocol = sckTCPProtocol Then
                m_Socket = SocketInit(AF_INET, SOCK_STREAM, IPPROTO_TCP)
                ' if we got a socket...
                If m_Socket <> INVALID_SOCKET Then
                    ' get the buffer sizes
                    If SocketGetOptions(m_Socket, SOL_SOCKET, SO_RCVBUF, lngBuffer, 4) <> SOCKET_ERR Then m_ReceiveSize = lngBuffer
                    If SocketGetOptions(m_Socket, SOL_SOCKET, SO_SNDBUF, lngBuffer, 4) <> SOCKET_ERR Then m_SendSize = lngBuffer
                    ' register messages
                    If WSAAsyncSelect(m_Socket, m_hWnd, SOCKET_MESSAGE, FD_READ Or FD_WRITE Or FD_ACCEPT Or FD_CONNECT Or FD_CLOSE) = SOCKET_ERR Then
                        Private_SetState sckError, "Bind", WSAGetLastError
                    End If
                End If
            Else
                m_Socket = SocketInit(AF_INET, SOCK_DGRAM, IPPROTO_UDP)
                ' if we get a socket...
                If m_Socket <> INVALID_SOCKET Then
                    lngBuffer = 1
                    SocketSetOptions m_Socket, SOL_SOCKET, SO_BROADCAST, lngBuffer, 4
                    ' get the buffer sizes
                    If SocketGetOptions(m_Socket, SOL_SOCKET, SO_MAX_MSG_SIZE, lngBuffer, 4) <> SOCKET_ERR Then
                        m_ReceiveSize = lngBuffer
                        m_SendSize = lngBuffer
                    End If
                    ' register messages
                    If WSAAsyncSelect(m_Socket, m_hWnd, SOCKET_MESSAGE, FD_READ Or FD_WRITE Or FD_ACCEPT Or FD_CONNECT Or FD_CLOSE) = SOCKET_ERR Then
                        Private_SetState sckError, "Bind", WSAGetLastError
                    End If
                End If
            End If
        End If
        ' start binding
        If m_Socket <> INVALID_SOCKET Then Bind = Private_Bind(LocalPort, LocalIP)
    End If
End Function

Public Function BytesReceived() As Long
    If m_Protocol = sckTCPProtocol Then
        BytesReceived = m_BufferSize
    ElseIf m_Socket Then
        SocketIOCtl m_Socket, FIONREAD, BytesReceived
    End If
End Function

Public Function CloseSocket() As Boolean
    Dim udtSockAddr As SOCKADDR_IN
    If m_State <> sckClosed Then
        Private_SetState sckClosing
        ' clear asynchronous process if active plus clean up the related memory
        If m_AsyncHost Then WSACancelAsyncRequest m_AsyncHost: m_AsyncHost = 0
        If m_HostentHandle Then GlobalFree m_HostentHandle: m_HostentHandle = 0: m_HostentPointer = 0
        ' now close the socket
        If m_Socket <> INVALID_SOCKET Then
            CloseSocket = (SocketClose(m_Socket) <> SOCKET_ERR)
            If Not CloseSocket Then
                ' something failed while closing the socket
                Private_SetState sckError, "CloseSocket", WSAGetLastError
            End If
        End If
        ' clean up
        m_Buffer = vbNullString
        m_BufferSize = 0
        m_LocalIP = vbNullString
        m_LocalPort = 0
        m_LocalPortBind = 0
        Set m_Request = New Collection
        m_RequestBuffer = vbNullString
        m_RequestID = 0
        m_RequestSA = udtSockAddr
        m_SendBuffer = vbNullString
        m_SendBufferSize = 0
        m_Socket = INVALID_SOCKET
        ' closed!
        Private_SetState sckClosed
    End If
End Function

Public Function Connect(Optional RemoteHost As String, Optional RemotePort As Long) As Boolean
    Dim lngBuffer As Long, lngIP As Long
    If m_Initialized Then
        ' remember given parameters
        If LenB(RemoteHost) Then m_RemoteHost = RemoteHost
        If RemotePort > 0 And RemotePort < 65536 Then m_RemotePort = RemotePort
        ' must have remote host
        If LenB(m_RemoteHost) Then
            ' create socket if we do not have one
            If m_Socket = INVALID_SOCKET Then
                ' create socket for that protocol
                If m_Protocol = sckTCPProtocol Then
                    m_Socket = SocketInit(AF_INET, SOCK_STREAM, IPPROTO_TCP)
                    ' if we got a socket...
                    If m_Socket <> INVALID_SOCKET Then
                        ' get the buffer sizes
                        If SocketGetOptions(m_Socket, SOL_SOCKET, SO_RCVBUF, lngBuffer, 4) <> SOCKET_ERR Then m_ReceiveSize = lngBuffer
                        If SocketGetOptions(m_Socket, SOL_SOCKET, SO_SNDBUF, lngBuffer, 4) <> SOCKET_ERR Then m_SendSize = lngBuffer
                        ' register messages
                        If WSAAsyncSelect(m_Socket, m_hWnd, SOCKET_MESSAGE, FD_READ Or FD_WRITE Or FD_ACCEPT Or FD_CONNECT Or FD_CLOSE) = SOCKET_ERR Then
                            Private_SetState sckError, "Connect", WSAGetLastError
                        End If
                    End If
                Else
                    m_Socket = SocketInit(AF_INET, SOCK_DGRAM, IPPROTO_UDP)
                    ' if we get a socket...
                    If m_Socket <> INVALID_SOCKET Then
                        lngBuffer = 1
                        SocketSetOptions m_Socket, SOL_SOCKET, SO_BROADCAST, lngBuffer, 4
                        ' get the buffer sizes
                        If SocketGetOptions(m_Socket, SOL_SOCKET, SO_MAX_MSG_SIZE, lngBuffer, 4) <> SOCKET_ERR Then
                            m_ReceiveSize = lngBuffer
                            m_SendSize = lngBuffer
                        End If
                        ' register messages
                        If WSAAsyncSelect(m_Socket, m_hWnd, SOCKET_MESSAGE, FD_READ Or FD_WRITE Or FD_ACCEPT Or FD_CONNECT Or FD_CLOSE) = SOCKET_ERR Then
                            Private_SetState sckError, "Connect", WSAGetLastError
                        End If
                    End If
                End If
            End If
            ' see if we not have a socket
            If m_Socket <> INVALID_SOCKET Then
                If m_Protocol = sckTCPProtocol Then
                    ' see if it is a string IP
                    lngIP = SocketGetIP(m_RemoteHost)
                    ' failed?
                    If lngIP = INADDR_NONE Then
                        ' we have to resolve the remote host...
                        If m_AsyncHost Then
                            ' cancel previous host query
                            WSACancelAsyncRequest m_AsyncHost
                            ' however, we keep using the previous memory block for HOSTENT
                        Else
                            ' get a new memory block
                            m_HostentHandle = GlobalAlloc(0, 1024)
                        End If
                        If m_HostentHandle Then
                            ' try to get the pointer...
                            m_HostentPointer = GlobalLock(m_HostentHandle)
                            If m_HostentPointer Then
                                GlobalUnlock m_HostentHandle
                                ' success, now we can resolve the host asynchronously
                                m_AsyncHost = WSAAsyncGetHostByName(m_hWnd, RESOLVE_MESSAGE, m_RemoteHost, ByVal m_HostentPointer, 1024)
                                ' success if we got that
                                Connect = m_AsyncHost <> 0
                                ' see which status we raise
                                If Connect Then
                                    Private_SetState sckResolvingHost
                                Else
                                    Private_SetState sckError
                                End If
                            Else
                                ' failed, free it up
                                GlobalFree m_HostentHandle
                                m_HostentHandle = 0
                                Private_SetState sckError, "Connect", GetLastError, "Failed to get pointer to allocated memory."
                            End If
                        Else
                            ' we failed to get memory for the asynchronous process
                            Private_SetState sckError, "Connect", GetLastError, "Could not allocate memory for asynchronous process."
                        End If
                    Else
                        ' we got a valid IP, try to connect!
                        Connect = ConnectIP(lngIP)
                    End If
                Else
                    ' UDP is a connectionless protocol
                    Connect = Private_Bind
                End If
            Else
                ' could not create socket
                Private_SetState sckError, "Connect", WSAGetLastError
            End If
        Else
            ' cannot start connection thanks to invalid settings
            Private_SetState sckError, "Connect"
        End If
    Else
        ' we do not even have all the requirements!
        Private_SetState sckError, "Connect"
    End If
End Function

Private Function ConnectIP(ByVal RemoteIP As Long) As Boolean
    Dim lngError As Long, udtSockAddr As SOCKADDR_IN
    If m_Initialized Then
        ' validate port
        If m_Socket <> INVALID_SOCKET And m_RemotePort > 0 And m_RemotePort < 65536 Then
            Private_SetState sckConnecting
            ' remember IP
            m_RemoteIP = RemoteIP
            With udtSockAddr
                .Addr = RemoteIP
                .Family = AF_INET
                .Port = SockethTons(m_RemotePort And &HFFFF)
            End With
            ' connect!
            ConnectIP = (SocketConnect(m_Socket, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
            If Not ConnectIP Then
                lngError = WSAGetLastError
                ' ignore "resource not available"
                If lngError <> sckWouldBlock Then
                    ' then imitate Winsock control
                    If lngError = sckAddressNotAvailable Then
                        Err.Raise lngError, "UniSock.ConnectIP", Private_FormatWSAError(lngError)
                    Else
                        Private_SetState sckError, "ConnectIP", lngError
                    End If
                Else
                    ' and thus we succeed
                    ConnectIP = True
                End If
            End If
        Else
            Private_SetState sckError, "ConnectIP"
        End If
    Else
        ' we do not even have all the requirements!
        Private_SetState sckError, "ConnectIP"
    End If
End Function

Friend Sub Friend_Request(ByVal hWnd As Long, RequestID As Long, SA As SOCKADDR_IN, Buffer As String)
    Dim udtSockAddr As SOCKADDR_IN
    If m_RequestID Then
        ' remember hWnd
        On Error Resume Next: m_Request.Add hWnd, CStr(m_RequestID): On Error GoTo 0
        ' copy there
        RequestID = m_RequestID
        SA = m_RequestSA
        Buffer = m_RequestBuffer
        ' clear here
        m_RequestID = 0
        m_RequestSA = udtSockAddr
        m_RequestBuffer = vbNullString
    End If
End Sub

Public Function GetData(Data, Optional VarType As VbVarType = vbByte Or vbArray, Optional ByVal MaxLen As Long) As Boolean
    Dim blnArray() As Boolean, bytArray() As Byte, curArray() As Currency, dtmArray() As Date, dblArray() As Double, intArray() As Integer, lngArray() As Long, sngArray() As Single
    Dim blnData As Boolean, bytData As Byte, curData As Currency, dtmData As Date, dblData As Double, intData As Integer, lngData As Long, sngData As Single, strData As String
    Dim lngFlags As Long, lngLen As Long, lngPtr As Long, udtSockAddr As SOCKADDR_IN
    ' which datatype to use?
    If VBA.VarType(Data) <> vbEmpty Then VarType = VBA.VarType(Data)
    ' which protocol we are using?
    If (m_Protocol = sckTCPProtocol) And (m_State = sckConnected) And (m_Mode = [Socket Binary Mode]) Then
        ' how much data to get?
        If MaxLen < 1 Then
            MaxLen = m_BufferSize
        ElseIf MaxLen > m_BufferSize Then
            MaxLen = m_BufferSize
        End If
        ' see if we get data at all
        If MaxLen Then
            ' now decide what to do
            If VarType And vbArray Then
                Select Case (VarType And Not vbArray)
                    Case vbBoolean
                        ReDim blnArray((MaxLen - 1) \ 2)
                        RtlMoveMemory blnArray(0), ByVal StrPtr(m_Buffer), MaxLen
                        Data = blnArray
                        GetData = True
                    Case vbByte
                        bytArray = LeftB$(m_Buffer, MaxLen)
                        Data = bytArray
                        GetData = True
                    Case vbCurrency
                        ReDim curArray((MaxLen - 1) \ 8)
                        RtlMoveMemory curArray(0), ByVal StrPtr(m_Buffer), MaxLen
                        Data = curArray
                        GetData = True
                    Case vbDate
                        ReDim dtmArray((MaxLen - 1) \ 8)
                        RtlMoveMemory dtmArray(0), ByVal StrPtr(m_Buffer), MaxLen
                        Data = dtmArray
                        GetData = True
                    Case vbDouble
                        ReDim dblArray((MaxLen - 1) \ 8)
                        RtlMoveMemory dblArray(0), ByVal StrPtr(m_Buffer), MaxLen
                        Data = dblArray
                        GetData = True
                    Case vbInteger
                        ReDim intArray((MaxLen - 1) \ 2)
                        RtlMoveMemory intArray(0), ByVal StrPtr(m_Buffer), MaxLen
                        Data = intArray
                        GetData = True
                    Case vbLong
                        ReDim lngArray((MaxLen - 1) \ 4)
                        RtlMoveMemory lngArray(0), ByVal StrPtr(m_Buffer), MaxLen
                        Data = lngArray
                        GetData = True
                    Case vbSingle
                        ReDim sngArray((MaxLen - 1) \ 4)
                        RtlMoveMemory sngArray(0), ByVal StrPtr(m_Buffer), MaxLen
                        Data = sngArray
                        GetData = True
                    Case vbString
                        strData = StrConv(LeftB$(m_Buffer, MaxLen), vbUnicode)
                        ' we make a string array simply by looking for line changes
                        If InStr(strData, vbNewLine) Then
                            Data = Split(strData, vbNewLine)
                        ElseIf InStr(strData, vbLf) Then
                            Data = Split(strData, vbLf)
                        ElseIf InStr(strData, vbCr) Then
                            Data = Split(strData, vbCr)
                        Else
                            Data = Split(strData, vbNullChar)
                        End If
                        GetData = True
                    Case Else
                        Err.Raise sckUnsupported, "UniSock.GetData", Private_FormatWSAError(sckUnsupported)
                End Select
            Else
                Select Case VarType
                    Case vbBoolean
                        If MaxLen > 2 Then MaxLen = 2
                        RtlMoveMemory blnData, ByVal StrPtr(m_Buffer), MaxLen
                        GetData = True
                    Case vbByte
                        MaxLen = 1
                        RtlMoveMemory bytData, ByVal StrPtr(m_Buffer), MaxLen
                        GetData = True
                    Case vbCurrency
                        If MaxLen > 8 Then MaxLen = 8
                        RtlMoveMemory curData, ByVal StrPtr(m_Buffer), MaxLen
                        GetData = True
                    Case vbDate
                        If MaxLen > 8 Then MaxLen = 8
                        RtlMoveMemory dtmData, ByVal StrPtr(m_Buffer), MaxLen
                        GetData = True
                    Case vbDouble
                        If MaxLen > 8 Then MaxLen = 8
                        RtlMoveMemory dblData, ByVal StrPtr(m_Buffer), MaxLen
                        GetData = True
                    Case vbInteger
                        If MaxLen > 2 Then MaxLen = 2
                        RtlMoveMemory intData, ByVal StrPtr(m_Buffer), MaxLen
                        GetData = True
                    Case vbLong
                        If MaxLen > 4 Then MaxLen = 4
                        RtlMoveMemory lngData, ByVal StrPtr(m_Buffer), MaxLen
                        GetData = True
                    Case vbSingle
                        If MaxLen > 4 Then MaxLen = 4
                        RtlMoveMemory sngData, ByVal StrPtr(m_Buffer), MaxLen
                        GetData = True
                    Case vbString
                        Data = StrConv(LeftB$(m_Buffer, MaxLen), vbUnicode)
                        GetData = True
                    Case Else
                        Err.Raise sckUnsupported, "UniSock.GetData", Private_FormatWSAError(sckUnsupported)
                End Select
            End If
            ' if we got the data and if not in peek mode...
            If GetData And Not m_Peek Then
                ' remove the data we got from the buffer
                m_Buffer = MidB$(m_Buffer, MaxLen + 1)
                m_BufferSize = m_BufferSize - MaxLen
            End If
        End If
    ElseIf (m_Protocol = sckUDPProtocol) And (m_State = sckOpen) Then
        ' peek mode?
        If m_Peek Then lngFlags = MSG_PEEK
        ' how much data to get?
        SocketIOCtl m_Socket, FIONREAD, lngLen
        If MaxLen < 1 Then
            MaxLen = lngLen
        ElseIf MaxLen > lngLen Then
            MaxLen = lngLen
        End If
        ' see if we get data at all
        If MaxLen Then
            ' now decide what to do
            If VarType And vbArray Then
                Select Case (VarType And Not vbArray)
                    Case vbBoolean
                        ReDim blnArray((MaxLen - 1) \ 2)
                        GetData = (SocketReceiveFrom(m_Socket, blnArray(0), MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = blnArray
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbByte
                        ReDim bytArray(MaxLen - 1)
                        GetData = (SocketReceiveFrom(m_Socket, bytArray(0), MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = bytArray
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbCurrency
                        ReDim curArray((MaxLen - 1) \ 8)
                        GetData = (SocketReceiveFrom(m_Socket, curArray(0), MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = curArray
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbDate
                        ReDim dtmArray((MaxLen - 1) \ 8)
                        GetData = (SocketReceiveFrom(m_Socket, dtmArray(0), MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = dtmArray
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbDouble
                        ReDim dblArray((MaxLen - 1) \ 8)
                        GetData = (SocketReceiveFrom(m_Socket, dblArray(0), MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = dblArray
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbInteger
                        ReDim intArray((MaxLen - 1) \ 2)
                        GetData = (SocketReceiveFrom(m_Socket, intArray(0), MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = intArray
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbLong
                        ReDim lngArray((MaxLen - 1) \ 4)
                        GetData = (SocketReceiveFrom(m_Socket, lngArray(0), MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = lngArray
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbSingle
                        ReDim sngArray((MaxLen - 1) \ 4)
                        GetData = (SocketReceiveFrom(m_Socket, sngArray(0), MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = sngArray
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbString
                        lngPtr = SysAllocStringByteLen(0, MaxLen)
                        PutMem4 ByVal VarPtr(strData), ByVal lngPtr
                        GetData = (SocketReceiveFrom(m_Socket, ByVal lngPtr, MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            ' we make a string array simply by looking for line changes
                            If InStrB(strData, CRLF) Then
                                Data = Split(StrConv(strData, vbUnicode), vbNewLine)
                            ElseIf InStrB(strData, LF) Then
                                Data = Split(StrConv(strData, vbUnicode), vbLf)
                            ElseIf InStrB(strData, CR) Then
                                Data = Split(StrConv(strData, vbUnicode), vbCr)
                            Else
                                Data = Split(StrConv(strData, vbUnicode), vbNullChar)
                            End If
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case Else
                        Err.Raise sckUnsupported, "UniSock.GetData", Private_FormatWSAError(sckUnsupported)
                End Select
            Else
                Select Case VarType
                    Case vbBoolean
                        If MaxLen > 2 Then MaxLen = 2
                        GetData = (SocketReceiveFrom(m_Socket, blnData, MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = blnData
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbByte
                        MaxLen = 1
                        GetData = (SocketReceiveFrom(m_Socket, bytData, MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = bytData
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbCurrency
                        If MaxLen > 8 Then MaxLen = 8
                        GetData = (SocketReceiveFrom(m_Socket, curData, MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = curData
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbDate
                        If MaxLen > 8 Then MaxLen = 8
                        GetData = (SocketReceiveFrom(m_Socket, dtmData, MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = dtmData
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbDouble
                        If MaxLen > 8 Then MaxLen = 8
                        GetData = (SocketReceiveFrom(m_Socket, dblData, MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = dblData
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbInteger
                        If MaxLen > 2 Then MaxLen = 2
                        GetData = (SocketReceiveFrom(m_Socket, intData, MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = intData
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbLong
                        If MaxLen > 4 Then MaxLen = 4
                        GetData = (SocketReceiveFrom(m_Socket, lngData, MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = lngData
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbSingle
                        If MaxLen > 4 Then MaxLen = 4
                        GetData = (SocketReceiveFrom(m_Socket, sngData, MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = sngData
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case vbString
                        lngPtr = SysAllocStringByteLen(0, MaxLen)
                        PutMem4 ByVal VarPtr(strData), ByVal lngPtr
                        GetData = (SocketReceiveFrom(m_Socket, ByVal lngPtr, MaxLen, lngFlags, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData Then
                            Data = StrConv(strData, vbUnicode)
                        Else
                            Private_SetState sckError, "GetData", WSAGetLastError
                        End If
                    Case Else
                        Err.Raise sckUnsupported, "UniSock.GetData", Private_FormatWSAError(sckUnsupported)
                End Select
            End If
        End If
    Else
        Err.Raise sckBadState, "UniSock.GetData", Private_FormatWSAError(sckBadState)
    End If
End Function

Public Property Get LineChange() As UniSockLineChange
    If LenB(m_LineChange) = 4 Then
        LineChange = [Socket CRLF]
    ElseIf m_LineChange = vbLf Then
        LineChange = [Socket LF]
    ElseIf m_LineChange = vbCr Then
        LineChange = [Socket CR]
    Else
        LineChange = [Socket Null]
    End If
End Property

Public Property Let LineChange(ByVal NewValue As UniSockLineChange)
    Select Case NewValue
        Case [Socket CRLF]
            m_LineChange = vbNewLine
        Case [Socket LF]
            m_LineChange = vbLf
        Case [Socket CR]
            m_LineChange = vbCr
        Case [Socket Null]
            m_LineChange = vbNullChar
        Case Else
            Err.Raise sckInvalidArg, "UniSock.LineChange", Private_FormatWSAError(sckInvalidArg)
    End Select
End Property

Public Function Listen() As Boolean
    Dim lngError As Long
    If (m_State = sckClosed) Or (m_State = sckOpen) Then
        ' first we must bind
        If Bind Then
            ' try to start listening
            Listen = (SocketListen(m_Socket, SOMAXCONN) <> SOCKET_ERR)
            If Not Listen Then
                ' failed
                lngError = WSAGetLastError
                Err.Raise lngError, "UniSock.Listen", Private_FormatWSAError(lngError)
            End If
        End If
    Else
        Err.Raise sckInvalidOp, "UniSock.Listen", Private_FormatWSAError(sckInvalidOp)
    End If
End Function

Public Function LocalHostname() As String
    LocalHostname = Private_LocalHostName
End Function

Public Function LocalIP() As String
    Dim lngLocalIP As Long
    If (m_State = sckOpen) Or (m_State = sckListening) Then
        LocalIP = m_LocalIP
    Else
        lngLocalIP = Private_LocalIP
        LocalIP = CStr(lngLocalIP And &HFF&) & "." & _
            CStr((lngLocalIP And &HFF00&) \ &H100&) & "." & _
            CStr((lngLocalIP And &HFF0000) \ &H10000) & "." & _
            CStr(((lngLocalIP And &HFF000000) \ &H1000000) And &HFF&)
    End If
End Function

Public Property Get LocalPort() As Long
    If m_LocalPortBind = 0 Then
        LocalPort = m_LocalPort
    Else
        LocalPort = m_LocalPortBind
    End If
End Property

Public Property Let LocalPort(ByVal NewValue As Long)
    If m_State = sckClosed Then
        If NewValue >= 0 And NewValue < 65536 Then
            m_LocalPort = NewValue
        Else
            Err.Raise sckInvalidArg, "UniSock.LocalPort", Private_FormatWSAError(sckInvalidArg)
        End If
    Else
        Err.Raise sckInvalidOp, "UniSock.LocalPort", Private_FormatWSAError(sckInvalidOp)
    End If
End Property

Public Property Get Mode() As UniSockMode
    Mode = m_Mode
End Property

Public Property Let Mode(ByVal NewValue As UniSockMode)
    m_Mode = NewValue
End Property

Public Function PeekData(Data, Optional VarType As VbVarType, Optional ByVal MaxLen As Long) As Boolean
    m_Peek = True
    PeekData = GetData(Data, VarType, MaxLen)
    m_Peek = False
End Function

Private Function Private_Bind(Optional ByVal LocalPort As Long, Optional LocalIP As String) As Boolean
    Dim lngLocalPort As Long, lngLocalIP As String, strIP As String, udtSockAddr As SOCKADDR_IN
    ' first of all check if we have bind already
    Private_Bind = (m_State = sckOpen)
    ' and if we have not...
    If (Not Private_Bind) And (m_Socket <> INVALID_SOCKET) Then
        ' determine local port to use
        If LocalPort > 0 And LocalPort < 65536 Then
            lngLocalPort = LocalPort
        Else
            lngLocalPort = m_LocalPort
        End If
        ' determine local ip
        lngLocalIP = Private_LocalIP(LocalIP)
        ' make a string too
        strIP = CStr(lngLocalIP And &HFF&) & "." & _
            CStr((lngLocalIP And &HFF00&) \ &H100&) & "." & _
            CStr((lngLocalIP And &HFF0000) \ &H10000) & "." & _
            CStr(((lngLocalIP And &HFF000000) \ &H1000000) And &HFF&)
        ' fill the structure for bind
        With udtSockAddr
            .Addr = lngLocalIP
            .Family = AF_INET
            .Port = SocketnTons(lngLocalPort And &HFFFF)
        End With
        ' bind it
        Private_Bind = (SocketBind(m_Socket, udtSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
        If Private_Bind Then
            ' remember IP
            m_LocalIP = strIP
            ' see if we had valid port
            If lngLocalPort Then
                m_LocalPort = lngLocalPort
                Private_SetState sckOpen
            Else
                ' get the local port
                lngLocalPort = Private_LocalPort
                If lngLocalPort <> SOCKET_ERR Then
                    m_LocalPortBind = lngLocalPort
                Else
                    Private_SetState sckError, "Private_Bind", WSAGetLastError
                End If
            End If
        Else
            Private_SetState sckError, "Private_Bind", WSAGetLastError
        End If
    End If
End Function

Private Function Private_FormatError(ByVal Number As Long) As String
    Dim lngResult As Long, lngStrPtr As Long, strError As String
    If Number Then
        ' initialize the buffer
        lngStrPtr = SysAllocStringLen(0, 65535)
        PutMem4 ByVal VarPtr(strError), ByVal lngStrPtr
        ' get the error message
        lngResult = FormatMessageW(&H1000&, ByVal 0&, Number, 0, lngStrPtr, 65536, ByVal 0&)
        ' if we got a string, place it in the string
        If lngResult Then PutMem4 ByVal VarPtr(Private_FormatError), ByVal SysAllocStringLen(lngStrPtr, lngResult)
    End If
End Function

Private Function Private_FormatWSAError(ByVal Number As Long) As String
    Select Case Number
        ' WSA_INVALID_HANDLE
        Case 6
            Private_FormatWSAError = "Specified event object handle is invalid."
        ' WSA_NOT_ENOUGH_MEMORY
        Case 8
            Private_FormatWSAError = "Insufficient memory available."
        ' WSA_INVALID_PARAMETER
        Case 87
            Private_FormatWSAError = "One or more parameters are invalid."
        ' WSA_OPERATION_ABORTED
        Case 995
            Private_FormatWSAError = "Overlapped operation aborted."
        ' WSA_IO_INCOMPLETE
        Case 996
            Private_FormatWSAError = "Overlapped I/O event object not in signaled state."
        ' WSA_IO_PENDING
        Case 997
            Private_FormatWSAError = "Overlapped operations will complete later."
        ' WSAEINTR
        Case 10004
            Private_FormatWSAError = "Interrupted function call."
        ' WSAEBADF
        Case 10009
            Private_FormatWSAError = "File handle is not valid."
        ' WSAEACCES
        Case 10013
            Private_FormatWSAError = "Permission denied."
        ' WSAEFAULT
        Case 10014
            Private_FormatWSAError = "Bad address."
        ' WSAEINVAL
        Case 10022
            Private_FormatWSAError = "Invalid argument."
        ' WSAEMFILE
        Case 10024
            Private_FormatWSAError = "Too many open files."
        ' WSAEWOULDBLOCK
        Case 10035
            Private_FormatWSAError = "Resource temporarily unavailable."
        ' WSAEINPROGRESS
        Case 10036
            Private_FormatWSAError = "Operation now in progress."
        ' WSAEALREADY
        Case 10037
            Private_FormatWSAError = "Operation already in progress."
        ' WSAENOTSOCK
        Case 10038
            Private_FormatWSAError = "Socket operation on nonsocket."
        ' WSAEDESTADDRREQ
        Case 10039
            Private_FormatWSAError = "Destination address required."
        ' WSAEMSGSIZE
        Case 10040
            Private_FormatWSAError = "Message too long."
        ' WSAEPROTOTYPE
        Case 10041
            Private_FormatWSAError = "Protocol wrong type for socket."
        ' WSAENOPROTOOPT
        Case 10042
            Private_FormatWSAError = "Bad protocol option."
        ' WSAEPROTONOSUPPORT
        Case 10043
            Private_FormatWSAError = "Protocol not supported."
        ' WSAESOCKTNOSUPPORT
        Case 10044
            Private_FormatWSAError = "Socket type not supported."
        ' WSAEOPNOTSUPP
        Case 10045
            Private_FormatWSAError = "Operation not supported."
        ' WSAEPFNOSUPPORT
        Case 10046
            Private_FormatWSAError = "Protocol family not supported."
        ' WSAEAFNOSUPPORT
        Case 10047
            Private_FormatWSAError = "Address family not supported by protocol family."
        ' WSAEADDRINUSE
        Case 10048
            Private_FormatWSAError = "Address already in use."
        ' WSAEADDRNOTAVAIL
        Case 10049
            Private_FormatWSAError = "Cannot assign requested address."
        ' WSAENETDOWN
        Case 10050
            Private_FormatWSAError = "Network is down."
        ' WSAENETUNREACH
        Case 10051
            Private_FormatWSAError = "Network is unreachable."
        ' WSAENETRESET
        Case 10052
            Private_FormatWSAError = "Network dropped connection on reset."
        ' WSAECONNABORTED
        Case 10053
            Private_FormatWSAError = "Software caused connection abort."
        ' WSAECONNRESET
        Case 10054
            Private_FormatWSAError = "Connection reset by peer."
        ' WSAENOBUFS
        Case 10055
            Private_FormatWSAError = "No buffer space available."
        ' WSAEISCONN
        Case 10056
            Private_FormatWSAError = "Socket is already connected."
        ' WSAENOTCONN
        Case 10057
            Private_FormatWSAError = "Socket is not connected."
        ' WSAESHUTDOWN
        Case 10058
            Private_FormatWSAError = "Cannot send after socket shutdown."
        ' WSAETOOMANYREFS
        Case 10059
            Private_FormatWSAError = "Too many references."
        ' WSAETIMEDOUT
        Case 10060
            Private_FormatWSAError = "Connection timed out."
        ' WSAECONNREFUSED
        Case 10061
            Private_FormatWSAError = "Connection refused."
        ' WSAELOOP
        Case 10062
            Private_FormatWSAError = "Cannot translate name."
        ' WSAENAMETOOLONG
        Case 10063
            Private_FormatWSAError = "Name too long."
        ' WSAEHOSTDOWN
        Case 10064
            Private_FormatWSAError = "Host is down."
        ' WSAEHOSTUNREACH
        Case 10065
            Private_FormatWSAError = "No route to host."
        ' WSAENOTEMPTY
        Case 10066
            Private_FormatWSAError = "Directory not empty."
        ' WSAEPROCLIM
        Case 10067
            Private_FormatWSAError = "Too many processes."
        ' WSAEUSERS
        Case 10068
            Private_FormatWSAError = "User quota exceeded."
        ' WSAEDQUOT
        Case 10069
            Private_FormatWSAError = "Disk quota exceeded."
        ' WSAESTALE
        Case 10070
            Private_FormatWSAError = "Stale file handle reference."
        ' WSAEREMOTE
        Case 10071
            Private_FormatWSAError = "Item is remote."
        ' WSASYSNOTREADY
        Case 10091
            Private_FormatWSAError = "Network subsystem is unavailable."
        ' WSAVERNOTSUPPORTED
        Case 10092
            Private_FormatWSAError = "Winsock.dll version out of range."
        ' WSANOTINITIALISED
        Case 10093
            Private_FormatWSAError = "Successful WSAStartup not yet performed."
        ' WSAEDISCON
        Case 10101
            Private_FormatWSAError = "Graceful shutdown in progress."
        ' WSAENOMORE
        Case 10102
            Private_FormatWSAError = "No more results."
        ' WSAECANCELLED
        Case 10103
            Private_FormatWSAError = "Call has been canceled."
        ' WSAEINVALIDPROCTABLE
        Case 10104
            Private_FormatWSAError = "Procedure call table is invalid."
        ' WSAEINVALIDPROVIDER
        Case 10105
            Private_FormatWSAError = "Service provider is invalid."
        ' WSAEPROVIDERFAILEDINIT
        Case 10106
            Private_FormatWSAError = "Service provider failed to initialize."
        ' WSASYSCALLFAILURE
        Case 10107
            Private_FormatWSAError = "System call failure."
        ' WSASERVICE_NOT_FOUND
        Case 10108
            Private_FormatWSAError = "Service not found."
        ' WSATYPE_NOT_FOUND
        Case 10109
            Private_FormatWSAError = "Class type not found."
        ' WSA_E_NO_MORE
        Case 10110
            Private_FormatWSAError = "No more results."
        ' WSA_E_CANCELLED
        Case 10111
            Private_FormatWSAError = "Call was canceled."
        ' WSAEREFUSED
        Case 10112
            Private_FormatWSAError = "Database query was refused."
        ' WSAHOST_NOT_FOUND
        Case 11001
            Private_FormatWSAError = "Host not found."
        ' WSATRY_AGAIN
        Case 11002
            Private_FormatWSAError = "Nonauthoritative host not found."
        ' WSANO_RECOVERY
        Case 11003
            Private_FormatWSAError = "This is a nonrecoverable error."
        ' WSANO_DATA
        Case 11004
            Private_FormatWSAError = "Valid name, no data record of requested type."
        ' WSA_QOS_RECEIVERS
        Case 11005
            Private_FormatWSAError = "QOS receivers."
        ' WSA_QOS_SENDERS
        Case 11006
            Private_FormatWSAError = "QOS senders."
        ' WSA_QOS_NO_SENDERS
        Case 11007
            Private_FormatWSAError = "No QOS senders."
        ' WSA_QOS_NO_RECEIVERS
        Case 11008
            Private_FormatWSAError = "QOS no receivers."
        ' WSA_QOS_REQUEST_CONFIRMED
        Case 11009
            Private_FormatWSAError = "QOS request confirmed."
        ' WSA_QOS_ADMISSION_FAILURE
        Case 11010
            Private_FormatWSAError = "QOS admission error."
        ' WSA_QOS_POLICY_FAILURE
        Case 11011
            Private_FormatWSAError = "QOS policy failure."
        ' WSA_QOS_BAD_STYLE
        Case 11012
            Private_FormatWSAError = "QOS bad style."
        ' WSA_QOS_BAD_OBJECT
        Case 11013
            Private_FormatWSAError = "QOS bad object."
        ' WSA_QOS_TRAFFIC_CTRL_ERROR
        Case 11014
            Private_FormatWSAError = "QOS traffic control error."
        ' WSA_QOS_GENERIC_ERROR
        Case 11015
            Private_FormatWSAError = "QOS generic error."
        ' WSA_QOS_ESERVICETYPE
        Case 11016
            Private_FormatWSAError = "QOS service type error."
        ' WSA_QOS_EFLOWSPEC
        Case 11017
            Private_FormatWSAError = "QOS flowspec error."
        ' WSA_QOS_EPROVSPECBUF
        Case 11018
            Private_FormatWSAError = "Invalid QOS provider buffer."
        ' WSA_QOS_EFILTERSTYLE
        Case 11019
            Private_FormatWSAError = "Invalid QOS filter style."
        ' WSA_QOS_EFILTERTYPE
        Case 11020
            Private_FormatWSAError = "Invalid QOS filter type."
        ' WSA_QOS_EFILTERCOUNT
        Case 11021
            Private_FormatWSAError = "Incorrect QOS filter count."
        ' WSA_QOS_EOBJLENGTH
        Case 11022
            Private_FormatWSAError = "Invalid QOS object length."
        ' WSA_QOS_EFLOWCOUNT
        Case 11023
            Private_FormatWSAError = "Incorrect QOS flow count."
        ' WSA_QOS_EUNKOWNPSOBJ
        Case 11024
            Private_FormatWSAError = "Unrecognized QOS object."
        ' WSA_QOS_EPOLICYOBJ
        Case 11025
            Private_FormatWSAError = "Invalid QOS policy object."
        ' WSA_QOS_EFLOWDESC
        Case 11026
            Private_FormatWSAError = "Invalid QOS flow descriptor."
        ' WSA_QOS_EPSFLOWSPEC
        Case 11027
            Private_FormatWSAError = "Invalid QOS provider-specific flowspec."
        ' WSA_QOS_EPSFILTERSPEC
        Case 11028
            Private_FormatWSAError = "Invalid QOS provider-specific filterspec."
        ' WSA_QOS_ESDMODEOBJ
        Case 11029
            Private_FormatWSAError = "Invalid QOS shape discard mode object."
        ' WSA_QOS_ESHAPERATEOBJ
        Case 11030
            Private_FormatWSAError = "Invalid QOS shaping rate object."
        ' WSA_QOS_RESERVED_PETYPE
        Case 11031
            Private_FormatWSAError = "Reserved policy QOS element type."
        Case sckBadState
            Private_FormatWSAError = "Wrong protocol or connection state for the requested transaction or request."
        Case sckGetNotSupported
            Private_FormatWSAError = "Property is write-Ba."
        'AIr."
        Case sckGetNotSupported
            Private_Forror = "In sckEp10069
            Private_FormatWSAError = "Disk quota exceeded."
        ' WSAESTALE
        Case 10070
            Private_FormatWSAError = "Stale file handle reference."
        ' WSAEREMOTE
        Case 10071
 requestedn

Public0"
       Private_FormatWSAError = "Disk quota exM   Case 10070
            Privatfile lEOBJ
 bUnicode)
                       OS shaping rate object."ce quota exM  OWSA_NOT_ENOUGH_MEMORY
      P300000000000000000000000000000000000000000tr(((ect."
     w        h         ssate_FormatWSAError = NCase sckGetNotSupported
owspec error."
        'T   ' O     Case 11005
            Pevate_Forwspec erropef_ocal ip
        lngLoi     Privat  Case 11005
            Pevate_Forwspec erropef_ocal ip
        lngLoi     Privat  Case 11005
            Pevate_Forwspec erropef_ocal ip
        lngLoi     Privat  Case 11005
      ratu lngLLLLLLLLLLLLcalwspec erropef_ocaw erropr And &pef_oca
End Functi, MaxLen)
  10071
 requestedn

Public0"
       Privablic Property Get LineChange() As Uni    OS 2              <O8or(sckUn yyoublic0"
       Privablic Propertpec erropef_ocal ip
    matWSAError = "Disk quota exceeded."
   t L(unctioI ct Case VarTy erropef_ocalec erro071
 requesor request."
     atError(ByVal Number As Long) As String
 0, Pp"Vali           Private_FowlU   End With
    ' WSAENOTase 11031
 005)
    c Proi0062
            Private_Foh
 equesr)iiiiiiiiiiiiiiiii sngArray
  I                 oortea        PrivaLLLLLLLLLLLLcalwspec erropef_ocaw erropr And &pef_oca
End Functi, MaxLen)
  10071
 requestedn

Public0"
       Privablic Property Get LineChange() As Uni    OS 2              <O8o  PrivaLLLLLLLLLLLLc"ock.LocalPort", Private_FoU
     erropef      Caf&) 000000000trPort > 0 And Local 000trPort p#iblicW      CStr((lngLocalIP And &HFF00&) \ &H10 Caf&) 0000oLLLLI If
    SCaf&L000000000000000esults."
 irPtre00esultsO connnnnnnnnnnnnnnnnnnnnnnnnnn dtmAerminwij SCaf0000peciic0"
       Privablic Propetpec erropef_oca     iic0"
       Privablic Prop ip
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA    ' WS     Privablic Prop ip
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA    ' WS     Privablic Prop ip
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA    ' WS     Privablic Prop ip
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA    ' WS     Privablic Prop ip
   AAAAAAAAAAAAAAAAAAAAAAAAAAAAAA     Privab)_FOUND
        Carropef_oca  2    g protoc3Private   ' WSAENOTSOCK
        Case 10038
            PrivatwPrivate )_FOUND
        Carrop  'APERATEOBJc3Private   ' WSAENOTSOCK
         CaseNOT_?R AENPrivate      WSAENOTSOCK
        CaetData Then
                              t== (Epped operation u_FOUNRIAe,perationr    ' WSA_QOS_NO_RECEIVERS
        Case 11008
            Pr Case=APE   P'APE   Privvvvvvvvvvvvvvvvvr)iiiiiiiiii7UAAAAAAAAAAa23
             "
        'AIr. Case 11007owlU   End WBalwspec erropa23 End a   Carrop  'APmatWSAErroAror = "DisONNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN Car23 EnAAA     Privab)_FOUND
        Carropef_oca  2    g protoc3PrPriopr A          Pr     eratiRND
   "R        oortea  r23 EnAAA  erved policy QOS eaNNNCeAAAAAAAAAAA     Privab)_FOmPrived policy QOS eaNNNCeAAAAAAAData 8     i4 e       ' WSAEACCES
        Case 10013
            Private_FormatWSAError = "Permission denied."
        ' WSAEFAULT
        Case 10014
            Private_FormatWSAError = "Bad address."
  EEEEEEETooooooooooooooooSAGetLastError
   Case 1L(,                     te_SetState sckError, "GetDaOEFILTERTYPatWSAError = "Be 100End   m_LocalPortBin m_Peek ThCase 10070
         fD_PET   ' WSAEACCESad polinnectioe not supported."
r."
        ' WSA_QOS_Etring
                        lngPtr =)                     Case vbDring
                        lngPtr =)                     Case vbDring
                        lngPtr =)                     Case vbDring
                        lngPtr =)                     Case vbDring
                        lngPtr =)               _FormatWSAEr70
  v                 ((((kr =)                    vab)_FOmPre 11026
   i23 EnAAA  erv    ElseIf MaxLen > lngLen Then
 S filting
    )_FOmPre 1108QOS eaNNNCeAAAAAAmati04vse    ElseYPE
4ess."
        ' WSnvalid QOS provider-specific flotrPort p#iblicWError = " connnnnnnnnnnnnnnnnnnnnnnnnnn dtmAerminQsnw2H   ElseIf  ' WSAENOBUFS
        Case 10055
 _ERR)
                        If GetData Tho                Case vbDring
 o      Case vbDring
          RLF]
    Elsebiiiiii 70
            Private0055
vate_SetSt     e vbDrier-specialPort = F   w        h          Private0055
vate_SetSt     e vbDrieDisk qu netData Th          Private", WSAGetor = "E          
  h    ER
        Case=ier-special9 =) (         Private0055
vate_SetSt     ebwLo      Case StSockAddr)) <        o WSAECONvbDrlseIf MaxLen > lngLen Then
 S filting
    )_FOmPre 1108QOS eaNNNCeAAAAAAmati04vse    ElseYPE
4ess."
        ' WSnvalid QOS provider-specific flotrPort p#ibmen RAAAAAAAAAAAAAAAAA    ' WS     Privabli       Cnnnnnnnnnnnnnnnnnn  ' WSAC  If GetData Tho  ie(ByVal Number As Long) As String
    SelerYYYYYYYYYCTw  Case sckGetNotr   CmdPrivate_FormatWSAError(lngError)
            End If
        End vting
)AError = "Too many open files."
        ' WSAEWOULDBLOCK
        Case 10035
            Private_FormatWSAError = "Resource temporarily unavaiovider-specific filterspec."
        ' WSA_QOS_ESDMODEOBJ
        Case 11029
            Private_:nge(fTcaw erroproviooption."
        ' WSAEPROTONOSUPPORT
        Case 10043
            Private_FormatWSAError = "Protocol not suppoOivate_FormatWSetSt     e vbSAER   ElsebiAError = "Resou35
  open files."
        ' WSAEWOULDr045
            Private_F = "Resou35
  od,op= vbCr
        Case [Socket NEi
        Case  Mode =Ke  ' WSTYPatWSAE   Caso_FormatWSAEr700  Private_F =&St  f "Resou35
ase 10054
 IAAAAAAAA    ' WS p       yVal Number As Long) As oPrivae(fTcaw ervate_For [Socket NEi
      XEACH
     ay
                        Elsef_oca     iic0"
                   y Then
                Select Case (VarTs [S  ' xivatP' xivatP' xivatP' xiva0oNQOS_ESDMODEOBJP resey Get LocalPort() As Long
    If mRAnd Local 000trPort p#iblicW      CStr((lngLocalIP And &HFF00&) \ &ct CaslPortB0    ' WS   6
         F00&) \ &ct Ccc xivatP' xivysAllvate_For [Socket NEi
      XEACH
     ay
                     fooooooooooooooooSAGet      ET_ERR)
       S.MaxLe2rtB0    ' WS  uS oPrivae(fTcaw eO       Case 10111
  Cargument."oror = "Operation now in progress."
 cD'xivatP' xivysA     Case 10047y object."
        ' WuivysA     CgPtr =) e 10FFIC_CTRL_ERROR
     Urivablic PrPrivaRR)
   ed cadWSA_QOS_ESDMODEOBJ
        Case 11029
ribbbbbbbbbbbbbbbbbbbbbP0t performedLFuS oPrivae(fTcaw eO       CicWErOPA of allc29
rsl_mMBW&HFF00&) \ &H10    Case 1() As Long
  rbbP0t performed    0t performedLFuS oPrivae(fTcaw eO  =B0      ' WSA_QOS_ESsste_Forror = "In sckEp1005
            Privatr = "In sc(fTcaw eO  =coPrivalpcaw e3  Err.Raise lngError, "U_ESsste_Forror = "In sckEp1005
            erropedtSockAdddilicropedpiiiiiiiiiiii1.atWSAError = "Cannot translate name."
 vf
End                    Private_FormatWSAError = Private_FormatWSAError = Private_FormESO   Private_FormatWSAError = Pr_RECOVERY
r allc29
rsl_mM iic0"
       Privab88vate_FormatWSAError =  AAlPort", Private_FoU
     erropef  =)        sAError         Privae_Fo0U
     erropef  OtWSA1' WSA_QOS_NO_RECEIVErr
    SA_QOS_RESERVED_P     ' WSA_QOS_ESsstLLLcalwspec erropef_ocaw erropr And &pef_Car2 AAlPort", PriI96
        atWSAError = "QOS generic err
   rPoIrro        Andi
End            ' WSA_QO ((((kr =_QOS_Enr     atWSAError =vatAEr        'T   ' OiRPrivab88vxLe2rtB0    ' WS  uS oPrivae(fTcaw eO    i0    iIsssssssss WS  uS oPrmber As Lc Pr'' OiRPrivab88vxLe2rtB0    ' WS  uS oPrivae(fTcaw eO    i0    iIsssssssss WS  uSi       Cnnnror         PrivaeLc PWLr)) <> S0Ctr = "In sc(fTcawg  Cnnnror         PrivaeLc PWssss WS  uSi       Cnc i0&now iCtr = "In s' rPoIrro        Andi
End      1' WSA_QOS_NO_RECEnow iCtr = "In s' xm     ' Oi uSi       Cnc     XEACH
     ay
  erYYYYYYYYYCe() AFecialDatSAER  d &Hrror, "GetTc in progress.or =sp      sAError         Privae_Fo0U
    tsr        'T   ' O  S.MaxLe2rtB0    ' WS  uS oPrivae(fTcaLr_RECOVERY
r allc29tAEr        'T   ' OiRPrivab88vxLe2rsssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssssCnnotr = "In sckEp1005
5CSockAddr, Len(udtSockAddr)) <> SOCKET_ERR)
                        If GetData ooooooSAReceiveFrom(m_Socket, dtm8  Case vbDring
          RLF]
 pef  OtW
       Hdtm8  Casperation now in progress."
 cD''''''eiveFrom(m_sssssssssssssssssssssssss   Carr &Hrro   Ca S.MaxLe2rtB0 Ca S.MaxLe2rtB0 Ca S.MaxLe2rtB0 Ca S.Mif GetD AAlPort", Private_FoU
     erropef  =)  ssssssssss   Carr &Hrro   Ca S.MaxLe2rtB0 Ca S.MaxLe2rtB0 CaitSAER  d &Hrrord        axLoaRMif GetD AAlPort", Private_FoU
     erropeYPE
4ess."  ' WSAENOTSOCK
        C= "Address already in use0trPort pf GetD AAlPort", Private_FoU
     erropefl7tWSAEAgsddress already in use0trnEAgsddress alre         e_Forlng
    ' OiRPrivab88vxL AAlnd000) & "." & _
            CStr(((lngLocalIP And &HFF000000) \ &H1000000) And &HFF&)
        ' fill the structure for bs al      drop7     EndERY
       
 o      C1000000) And &HFF&A flowAPort",h000)      Privae_Fo0U
    tsr        'T   ' O l ssssssd    flowAPortansaction orIllc29
rsl_mMBW&HFF00&) \ &H10    Case 1() As Long
  rbbP0t pecti     'T   ' O"Ti== "Nonauthor7svP' xiiiiiiiiiiiiD        foooooooooSA_QOErsl_mae_Fo0U
    tsr        'T   ' O l ssssssd    flowAPortansactiesult = 5g
 ab88vxL AHFF00&) \ &HoaRMif GetD AAlPort", Private_FoU
                 bICval9(e8dSrmay         Private_Form ii7UA genericfata ooooD        foif2i:ange a
        oooSAReceiveFrolG<Ve As OiRPrivU=r a
        ooo
   lter style."
        ' WSA_QOS_E      ' WSA_QOS_E      ' Wg
  rbbP0t pecti     'T   ' O+ styfooooo Case vb     Prie flowAPor9 =) (   dda
        Else
     PORT
        Case 1raEiTo   0e_Fo, i0   )oU
flowAPor9 =) ablicmore rLe2rssssssssssssss       i fooi   AlowAPor9 ISEE    PrivatiR
        Case 11014
           orIllc29tD AAlPort",      fooooosh.LocalIP An0LocalIP An0LocalI   CaserMtWSAErroreAlowAPor9 ISvate_FoU     E          bICval9(e8dSrmay      YYYYYYYYYCTw  C         "
        ' WSA_QOS_TRAFFIC_CTRL       Else
     PORT        bILLLLcalwR  g   ' Wg
  rfnSAErrorssss    t   Pevate_Forwspec err&rorsCStr(((lngLocalIP And &HFF000000=) lPortp00000=) lPortp000A_QOS_E      ' Wg
  rbbP0t ppppppppppppppppppppppphp    Cnc      10111
YYY   ' fill thm     ') <> SOCKET_ERR)
            cific 7WBalwspec e    10111
YYY   ' fill thm     ') <>  '         cif cific 7WBalwsp'     End If
        e ') <> SOCKET_ERRti     'T   _mM iiiiiiiiiiiigormatWSAEr700  PrM iiiib7y=" cific 7WBalwspec b7y=" cific 7WBalwspec b7y=" wAPortansacties' WSA_QOS_TRAFFIC_C)
            End If
        End Ift pecti     way      YYYYYYYYYAsu 4 ' n If
        End Ift pecti     way    (e8dSrma" cifit< N  ' WSA).LineChan11005
   FoU
     erropefl7tWSAEAgsddData", WSERR)
            cific ase 10043
            Private_FormatWSAError = "Protocol not s FoU
0047y object."
 Etring
                        lnging
 0, Pp"1bP0t pppppppppppppppppppngData,   00000
    WSERR)
epeAAAAAAAAte_FormatWSAErr    f0t pppppivatergData,   00000
    WSERR)
epeAAAAAAAAte_Foro4k   i0    iIsssssssss "Protocola,   000iveFrom(m_Socket, curData, MaxLen, :) <> STf    DatafAAAAAte_rivae(fTcaw eOss "f STf AAAAA  MaxLen, :) <> STf    DatafAAAAAWBalwsp'     End If
 b88vxL AHFF00&) \ &HoaRMifivate           oaRMifivate    rorcmily = AF_INET
            .Port = 
 b88vxL AHA).LineChan11005
   FoU
     erropefl7tWSAEAgsddData", WSERR)
            cific ase 10043
            Private_FormatWSAError = "Protocol not0iveddData", Wppefl7tkOpenC = 5g
 ab     E Ca S.MaxLeeeeeeeeeeeee:]fiR)
, WSEction reset by peer."or = "Procedure call table is invalid."
7Data", WSERR)
    rorcmily = AF_INET
            .Port = 
 b88vxL AHA).LineChan11005
   FoU
     erropefl7tWSAEAgsddData", WSERR)
            cific ase 10043
            Priva            cific ase 10043
t       cific ase Wsm_Socket, cutype."
        ' WSA__________________nnnnnnnn dtmAerminQsnw2H   ElseIf  ' WD fS                 cific ase Wsm_Socket, c    Case 11013
                       a", Wpp_Forific ase 100rmatWSAErrP     ____EdCval9(e8dSrmay      YYYYYYYYYCTWsm_SWSAESTALE
       Dg    arP     ___ PriaWSERR)
   Case 10044
    013
                       a", Wpp_Forific ase 100rmatW
            CStr(((lngLocaData, MaxL=", Wpp_F    bICval9(e8dSrmay       a", Wpic asARAMETERI        RECEIVErr
    AAAA
       
   oe_FormatWSWse    rorcmilywij SCaf0000peciic0"
e8ddState)
atwPrivky Gewspec b7y=R)
            cif OS 2              <O8or(sckUn yyoublic0"
L      foif2i:ange a
          ') <>  '     dp0iveddData", Wppefl7tkOpen'   'ocke 
  ge a
          ') <>  '     dp  =)  ssssssssss   Carr &Hrro   Ca S.MaxLe2rtB0 Ca S.MaxLe2rtB0 CaitSAER  d &HAo 'ocke 
  ge a
      O   <O8or(scf OS 2         AHAO8or(scf     ' WSnvalid 8vxLia              0lace nnnnn AEN Wpic asARAMEB0 Ca S      a", Wpi a", ' WD fS                 cific a       ssss   a              0lace nnnn)p     cific a       ssss    fS        fWse  cifit   W      CSts fS                 cific a    Re
        If LocalPort > 0 And Localdwspec b7y=R)H 10009
               C     O   <O8or(scf OS 2       S  ' WSnvalid 8vxLia              0lace nnnnn AEN Wpic asARAMEB0 Ca S      a", Wpi a", ' WD fS                 cific a       r1lc29ance "Call wa   r1ror peciSEirrivU=r a
        ooo
   lnce EB0 Ca S      a", Wpir8r = "Lro   Ca S.MaxLe2rtB0Data(e8dSrBdSrmay       a", valid."
        ' WSAEINVALIDPROVIDER
        Case 10105
   c asAALIDPROVIDER
     PIet, cO$r(lngmatWnfse 11019
            Private_FoDud       .Port = SocketnTons(le 10105
   c asAALIDPROVIDER
  L  ciuoF33333333s                     OSifit   W1Vro   Ca S.Ma  W1VrretnTons(le 10105
   c asAALIDPROE             eciSEirrivU=r a
        ooo
   lnce EB0 Ca S      a", Wpir8r = "Lro   Ca S.MaxLe2rtB0Data(e8dSrBdSrmay       a", valid."
        ' WSAEINVALIDPROVIDER
        Case 10105
   c asAALIDPRO  c asAA"
      sCAeineCS.MaxLssssssssssssssssss,5
   DPRO    Case      ir8r = "Lro   Ca   ir  PrirStr(((lngLocaO  c asAA"
ate           oaRM  Case vbDA_OPERATION_ABORTED
        Case 995
 rbbP0t performed    0t perN Wpic asA Cas=INVALIDPROVIDER
alIP An0LocalIP An0LocalI   CaserMtWSAErroreAlowAPor9 ISvate_FoU     E        rN Wpic asA Cas=INVALIDPROVIDER
alIP An0LocalIP An0LocalI   CaserMtWSAErroreAlowAPor9 ISvate_FoU     E        rN Wpic asA Cas=INVALIDPROVPa", WpiiiiiinalIP    Private_FormatWSAError = "Destination addr      Else
    asAALIMif GetD AAlPoin4ssse_FoU
     erropef  =)        sAErro   Else
 OS_EG    PeekDats     a", valid."
          t== (Epped ope1Pi a", ' WD fS        VIDERr = "Invalid oivate_       VIDERr = "Invalid oivapetpec error = "Incorr     rirSf LocalPort > 0 Anas=INVAOneChan11005
   FoU
     erropefl7tWSAro   Ca ta(e8dSrBdSrid oivapeCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC10111
 CCCCCC oaa         cie 11008
        Case 1101
   c ascie 1CCCCCCCCCCCy   c asAALIDPROVIDER
    CCCCCCCCCCCCC+CCCCCCCCCCCCCCCCCCCCCCCC         ECCCCCCCCCCCCCCCCCCCCCCCCCCr = "CCCy   c aspck.GetData   cieR
    s
      sCAeineCS.0005
   Dopef_oca     iic0"
      r    ng
      f       teOREMOTE
        Case 10071
 r     e    10111
YYY   ' fill thm  An uS oPrYror = "    Case 10071
 r   O
        Case 10071
 r  ormatD      f   Mse 10071
 r  ormatD      f   Mse 100&HFV1
 r)i"InwPriCif2i:ange a
        ooose 10071
    ooose 10071111tkOpenC = 5g
 ab     E Ca S.MaxLeeeeeeeee     End If
            End If
 d0h0C = 5g
 ab     E Ca S.Man(lngSsi:ang  cif OS 2       OREMOTE
                                                       CCCCCCCCCCCCC7= "Invalid oivate_       VID_FormatWSAError = "Connection timed outsAAlPort", Private_FoU
     erropeYPCCCCCCCCCCCCr = "CCCyS= "Incor   Privvvvvvvvvvvvvvvvvr)iiiAUINVALIDAd oivate_   PeekDats    *CCCCCCCCr = wWmU
     *CCCCCCCCr = wWmU
     *CCCh    U1007TIf
lowAPorCCr = CCCCC7= "Invalid o  L  Invali    PriebSW
    013-CDo, ' WD fS  GeCCCr = M
        Case sckGetNotSupruH   erVAOneChan11> <O8B 7WBalwspeckr =_QOS_Enr     atWSAError =vatAEr   71
    ooos  013-CDo, ' WD fS ' WSnvalid 8hckInvalidArg,
 IAAAAAA        rmatWSAError = Private_FormESO   PWSAError(lngError)
 f       teOtInvali     ' WSANO_RECOVERY
        Case 11003
      ruoF33333333s     Lli   tocoSaction orIllc29
rsl_mMBW&HFF00&) \\\\\\\\\\\\\\\\\\\\\\\\\\    r)
 f   P orIllc2     'T   ' O+ styfooooo CaEnd If
        e ') <> SOCKEError = "DestinaSorIllc2     'T   '  tocoSaction orIllc29
rsl_mMBW&HFF00&) \\\\\\\\\\\\\\\\\\\\\\\\\\    r) o fD_PET   port
     ' WSASYSedtmAion orIllc29
rsl_mMBW&HFF00&3    Fd0rIllc29
rsl_SOCKrIllc29
rsl_S fS  f    Case sckkkkkkkk        PrivaeLc PWLrA                VIDkkkkkkkkSYS'       e ') <> SOCKEErl_SOCKr(le 1a=lic Pr N Wpic asSw                ' WSASYSedtmAion orIllc29
rsfa", WSERR)
kSYS'       e ') <>iorIllc29
 5
      lV\\\\\\\\ac29
 5YS'       e ') <>iorIllc29
 5
      lV\Gcific ase
    )    eOCKr(le 1a=lio Privat ') <> SOCKEEion orIllc29otWSig,
 IAAAAAA      End IfIllc29
 5
      lBlN  *CCBe resoBLEr   7Oo  Case 995
 rbbP0t perform     lBlLEr   7Oo  Case 995
 rbbpef      Caf&) 0000000CCr = wWmfr)NOTSOCK
  (EvalidrlHgErre_FormESO   PWSAError(lngELocaO  f   Mse 100&HFV1
 'pe 11019
    rt
     ' sllc2)NOTS     E
             te_    14Be resoBLEr   7Oo  CAError(lngELocaO  f   Mse 100&HFV1
 'Ert pfK Case Rrvice type  foif2i:ange a
 ormatD      f   Mse CaMI An0LocalIP An0LocalI  ng  ooos  013      e ') <>iorIllc29
 5
      lV\Gcific ase
  hE     e ') <>iorIlt p
    ' WSAENOTase 11031
 005)
 n)     lV\GcificaiPrivvvvv1   lV\GcificaiPri  asAALIMif GetD AAlPoin4ssse_FoU
     erropef  =)        sAErro   Else
 OS_EG  *CCCCCCCCr = wWmU
     *CCCCCCCCr =sse_FoU
     e     D 1a=lielI  ng  oo =)        Ca c=)   ,iiinalIP    Private_FormatWSAError = "Destination addr95
 rbbP0t performed    0t perN Wpi            Priv
    Wpi            Priv
    Wpi      "Des8hckInvasnmU
    Son addr95
 rbbP
 OS_EG  *CCCCCCCCr = wWmSaOocalIP An0LocalI     lowAPor9 =) ablicmor          ific FormESO 
YYY    
  aswWmSa   Wpi   ,   (FO,0ckInvasnmU
    Son addr95
 rbbP
 OS_EG  *CCCCCCCCr = wWmSaOocalIP An0LocalI     lowAPor9 =) ablicmor          ific FormESO 
YYY    
         Private_C    With r)NOTSOCK
            ific FormESO 
YYY  Private      WSAENOTS     ifiPicmor          ific FormESO 
YYY    
 OermatWSAError = "QO0  ifiPicmor         addr9D AAleO 
YYY    
 OermatSe
  IhBlLEr   7O     Private_FormatWSAError = "QOS no ilI     lowAPor9 =) ablicmor          ific FormESO 
YYY    
         Private_C    With r)NOTSOCK
            ific FormESO 
YYY  Private      WSAENOTS     ifiPicmor r&) 000000  Prrivat Pri _DATlc2AerN Wpi   c cificWmfr _DATlc2AerNrlsssss>fs10105
   c asAALIDPROVIDER
  L  ciuoF33333333s          rmatSe
  IhCa S.MaxLe2rtB0Daa Son AAAAAAAAAAAAAAAA   CCCCCrror = Priva   
         Private_C    Ri:CfS        AErr
            cific 7WBalwspeceA =) (         Private0055
vate_SetSt     
vate_      ' WSAEUSERS
   ' WSAEUSER,C   ific 7WBalwsIu ciuoF33333R&A   e ') <>iorIlt p
 ific 7WBalwsIu ciuoF33333R&A   e ') <>iore
YYY    
nC,
         o
   lnce EB0 Ca S   YormESO 
Yesor request."5
 rbbP
 ivate_FormatWSAError = "SaOocalIP An0LocalI     lo"QOS no ilI     lowAPor9 =) ablicmor  
 ivate_FormatWSAError = "SaOocalIP An0LocTmfr _DATlc2AeSe reference."
       P00000ise 1raEi
 ineChe(le 1a=llllllllllllly' Priva   
         Private_C    Ri:CfS        AErr
            cific 7WBalwspeceA =)         <O8or(sckUn svate) (         Privat     ci Els  
        Dific 7WBa     <O8oA <>  '     dp0iveddData", Wppefl70_RECEnow iCtr =  c=)   ,iiinalIP    Private_FormatWUn n so005)
 n)     lV\GcificaiPrivvvvv1   lV\GcificaiPri  asAALIMif GetD AAlPoin4ssse_FoU
     erropef  =)        sAErro   Else
 OS_EG  *CCCCCCCCr = wWmU
     *CCCCCCCCr =sse_FoU
 uiE1
 r  oVngLocalPorcalIP An0    C=)  Re_Format     EnRU
     *CCCCCCCCr =sse_FoU
 uiE1
 r lIP e 10r =ssCEnow i7WBalwsIu cxu FormESO 
YYY    
         Pri?no ilI     lowAPor9 =) ablicmor          ific FormESO 
YYY <0sSO 
YYY <0sSO 
YYY Y    Case 10071
 request
         cxu FormESO 
YYY    
         Pri?no ilI     lowAPor9 =) lAAAAmati04vse    ElseYPE
4ess."
   A =)        <O8or(sckUn svate)
YYY    
    pecialPocmor       ,uiM     Flial9 =) (   04vse    Elsor(sckUn svate)
YYY    
    pecialPocmor       ,uiM     Flial9 =) (   04vse    b1objectnauthorrrrrd_FoU
     er   
    pecialPok= "SaOocalIP An0LocTmfr _DATlc2AeSe reference."
       P00000ise 1raEi
 ineChe(le 1a=llllllllllllly' PrilPorcalIP erNrps'tnLe2r IhCa S.MaxLe2rtB0Daa Son AAAAAAAAAAAAAAAAAAAA n)     lVdressaa S
   0ise 1raEi
 ineChe(le 1aIWSERR)
            cificN  ormatD      f   Mse 100&HFV1
 r)i"InwPr22r I   ' WSr    lVdressaa S
tp00000=l ooos for the 
   eric error."
,so005)YYY <0sS  c   P000ocmor      -CDo, ' WD fS  GC0u(      Pri?no il EC10111
 2222222222222222222222222B
YYY    
    peci"InwPr22r I R,C   ific 7WBBBBBBBBBBBBBBBB(1000000) And &HFF&)
) &HFFdc 7WBBBBBBBBBBBBBBBB(1000000) An peci
     CCCCCCCCr = wWmU
  1011Private_ForI1
 2BB(100000=lllllllllldr)) Len(udtSockA      isS  c  dc 7W    r    ng Wsm_Sockee sc C=) \W'fokGetNotSupported
            Pce "QOS traffic control error."
  ysm_Socket, cutype.S t    Ipd oi    tvate_C    W  Ipd oi  SicaiPri  as ' OinOcutype.S tf_ocaw erropwcutype.S tf_ocaw erropwcutype.S tf_ocaw erPoIrro               <O8orP1doc0061
            Ptype.. object."
    d oi    tvate_C    W  r(C=)  Re_Format     EnRU
   A   s
          n0061
            Ptttttttttte_ForI1
 2BB(100000=lllllllllldr)) Le   P00000ise 1riva   
       prror = "       0GetD AAlPo(oe.. object."
    B(100rmatSe
 r)i"InwPllllllllse lngError,  Yx00ise 1riva   
       prror = N     Case 10070
            Private_FormatWSAError = "e
 r)i"InwPllllllllse lngError,  Yx00nC = 5g
 ab     E Ca S."InwPr22r    
    peco#iblicW       cifiH(ormESO 
YYY tneChe(T Ip  Case 1007U
   A   s
        ocTmfr _    lo"QOB   n0rmESO 
YYY t          Ppeco#if       'Lhe(T SO 
oc Private_FormatWSAError = s
 ocTmfr _    lo"     ESO 
YYY t          Ppeco#if       'Lhe(T SO 
oc Private_FormatWSAError = s
 ocTmfr _    lo"     ESO 
YYY t          Ppeco#if       'Lhe(T SErroPpeco#ifPrivate_Format ise 1r _  cormat ise"e
((((((((((((((((((((ERS
       bV    ssssssd oPpe           (_he(T SErroPpeco#ifPrivate_Format ise 1r _  corm((ERS
  =atWSAError = "Network drror = "Network drror = "Netwo6o#ifPrivate_Format ise 1r _  cormat ise"= s
 ocT  i2F  <>iorIlt p
    'D_  cormat ise"e
((((((((((((((((((((ERS
   'T           opef  OtWSA1' WSA_QOS_NoPpe           tr(or = "NeiAError = "ReWppefl70_RE4t     EnRU     lV\Gcwirror , Pr'' Oi((((((((ERS
   'T              WMSA_QOS_NoPp    iffDReWp/rYYYYYYYYYCTYPatWSAE   Caso_FormatWSAEr7Cssssssss\Gcwirroc C= PrivC\GcOe
((((AAAA '     dp\GcwNotSupported
    y/rYYYYYYYY\GcOe
((((AAAA '((((vate   ' W rivate   ' WSAENOTSOCK
        Case   Date_C    W  r(C=)  RswNotSupported
    y/rYcormat ise"La3gY
 5YS'lo"   Supported
 C\GcLPrivate"La3  10 1otSupporYYYYYDATlc2Ae         ific FormESO 
r         Sported
    y/rYcor22r I R,      ocTmfrB y/rYcorma   ifiCCCCC PrivGcLPrivate"La3  10 1otSupporYYYYYDATlc2Ae         ific FormESO 
r         Sported
10 
            Private_Fodported
    y/rYYYYYYYY\GcOe
((((AAAA FifiH(ormE    S    ESO 
e   P   Ipd
    y/r    YYDATlc2Ae  YYYYH       (_he(T e    101pi       cormat ise 1riva   
         ' W fic control erre    101pi       cormat ise       Case  lc2Ae  YYYYk  YYY YYYYH          SOCKrIlS  <O8or(scf OS 2         AHAO8or(scffffffffffffffffffffffffffffff    
  1011PrivOS 2 fffffDS     AHAO8or(scfffffffffffffffffffffdgPtr =)AAAte_Foro4bdr      cifi         ific FormESO 
YYY    
  aswWm(
         fffDS  ffffdgPtr =)AAAte_Foro4bdr      cifi      =,firme222221oOivate_F    aswWm(
 rmatWSAErros' WSEr7CsssssssErros' WSEr7CsssssssErro  ffesV
 IAAA ocTmfrB      ' W fic controlQOS no ilI     lowAPOtW CCCCC  lowAPCacTmfrB    DnrIlt p
    'D_  cormat ise"tB    os' WSEr7Cssss   cifi     mat ise"e
Td
    y/rY30000000000Y30000000000Y30000000000Y30000000000Y30000000000Y30000000000Y30000000000Y30000000000Y30000000000Y30000000000Y30000000000Y30000000000Y30000000000CP cifi   aswW0000000000YT   cifi         ific FormE00Y30000000000CP cifi   aswW0000000000YT0iFormat 000000Y3000000000f      ,iiinade.S tf_ocaw bsErros' WSE0Y30000000000CP c   y/ropef      C,h Date_C    WivU=r a
   rswW0000000000YT0iFormat 000000Y3000000000WSA_QO
        ' WSnvalid QOS provider-spe)c Proyaw bs Car     'S _  cormat ise"e
(((((H  y/ropef     sssssssErros' WcifisdWSnvalid QOS pi   A       Case-initialize."
        ' WSASYSCAfic contric Forme Car     'S _  cormat*formed    0t perN Wpic asA Cas lV\G0Y30000000000Y30000000000000000000000000000000P_FoU
 uiE1
 r lIP e 10r =ssCEnow i7WBalwsIu cxu FormESO 
YYY  lwspeckrxu FormESO 
YYY  lwspeckrxu FormESO 
YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYsivOS 2WSAErrllllllllll   lizeYYYYYYYYYYYYYYYYYYYYYYwiE1
 r  oVngLocalPo(objectnauthorrrrrd_FoU
     er   
    pecialDg    arific flotrPort p#ib000i
    pecr22r doYYYYY er   
 ip
    mat)  RswNotSupported
    y/rYcormat ise"La3gY
 5YS'lo"   Supported
 C\GcLPrivate"La3  10 1otSupporYYYYYDATlc2Ae         ific FormESO 
r         Sported
    y/rYcor22r I R,E0000000000Y30000000000Y30000000000CP cifi   asa S   YorwA    Slt erropefl7tWSA erropefl7tWSA erropefl7 Sported
    y/rYcoEotr   CmdPrivate_FormatWSJhrYYYYYDATlc2cyC    Case   Date_C  o
oc Pted
drminwij SCafTrYYYYYDATYYYY er   
 ip
    mat)  RsW0000000000YT0iFormat 000000Y300iFormatw i7DATlc2cyCYYYYYYYsivOS 2W   p
    mat)  RswNotSupported
    y/rYcormat ise"La3gY
 5YS'lo"   Supported
 C\GcLPrivate"La3  10 1otSupporYYYYYDATlc2Ae         ific FormESO 
r         Sported
    y/rYcor22r I R,E0000000000Y30000000000Y30000000000CP cifi   asa S   YorwA    Slt errfl7tWSA erropefl7 Sp0000000000Y300000     cor   Pr   Case e30000orwA    r,   INET
        me"La3    SOCKrIlS  <O8or(scf OS 2   AError(lngError)
 f       teOtInvali   YYY er 05
rYcor22r I R,E000000000cs0000Y300e4ess."
ilI     lor)
 f       teOtInvali   YYY er f       teOtInvali   YYY eEA05
rmatWSAorYYYYYDATl4ess."
ilI  or)
 f 2         AHAO8or(scf     ' WSn   CmdPrivate_FormatWS000000000000Pted
drminw0000000000Pted
drminw00000000007tWSAo5t0D&3   ,iiinalIP eEA(AO8or(scf   mf       teOtInvali   YY        .PoruForwspecocalPort", Ne_FormatWSAError = "ormafffffuAop= "SylBlNnalIP eE  t0n   ' 10r id QOrTlc2cyCYYYYYYYsivOS 2W   p
    mat)  RswNotSupported
    y/rYcorase eF00&) \ &ct  open files."
        ' WSAEWOULDr045
nOtIid QOrTlc2cyCYYYYYYYsivOS 2W   p0000000YT   cifi         ific U"ormafffffuA ilI     'T   ' OiRPrivilI     'T   ' OU
        ' WSAEWOULDr045
nOtIid QOr    ' WSAEwr)) e0055
vorted
onrted
    y/UUUUUUUUUUUUU0000000o performed."
        ' WSAEDISCON
  nOtIid QOr  atWSJhrYYYEwr)) e00pNrivate   ' WSAENOTSOCK
 333333s     
drminwted
iiiii erropef  =)  sIali oSA_d4Ig
 tnYY    
         PrivtsLen(udtSoYYYYYDLMaxLen, :) <> STf    DatafAAAAAte_rivaeeeeeeOd      ific U"sckEp1005
5CSockAddr, Len(udtSockAddr)) <> SOCKETC6o"
ilI     lor)
 f   <> STf    i32
          id QOrTlc2cyCYYYYYYYsivOS 2W   p
  rmatWSAError = "Wrong pr   p
  rmatWSe    i3lIic U"si
EndO       CicWda S   i32
        lAe                    fTrYYYYYDDDDDDD_C    tc2cyCYYYYYYYCmdPrivate_Formapeco#if       'LhN         fTrYYYsA
 f j2
EndOAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA    tc2cyCYYYYYaatWSSSSSSSSSSSSSSSSSSSSSSSSSHllllllllllllll"A00000000Y30itAAAAAAAAAAA    ma lllllfir "Network drror = "      ' WSYDLMllllllllllltd QOSSSSSSSSSSSSSSSSSSHllllllllllll0 
            Private_Fodported
    y/o0000Y30itAAAAAAAAAAA    ma lllllfir "Network drror = "      ' WSYDLMupporl' OiRP      Case
    y/o0000Y30itAAAAAAAAAAA   aa    ' WSYDLMupporl' OiRP        ific", Ne000Y30r            Private_Fodported
    y/r(f Merropefl7 Sported
r = "ormafffffuAop= "SylBlNnalIP7O     Private_FormatWSAEr
   AAAAAAAAAAAAAAAry/r(f Merropefl7 Sported
r = "ormapporl' OiRP        lfir "Network drrorose xA.ormainw00OWNPf ific",     lfir "Neprl' OiRP   ' WSAEwr)) e0055
vorteyd
r = "ormapporl' ,o drror = i n30000000000CP c   y/ropehSSSSSHllllllllllllll"A000Tlc2Ae ificDAA    ma lllllfir "Networkoi006AErr  ' W fic control eed policy QOS=  c=)   ,iiinad
    y]ni n300000000Y30000000000Y300)Uted
Lrrl' oDr045
nOtIid QOr    ' Wat      ' WSA_llllllma lllllfir "Networkoi006AErr  '     Private_F = Olllma  ln  cifi  fffDS  ffffdgPtr00000rntc2cyCYYYY1011 If GetData Tho   lfir "29
srp
  sO       id QOse Ptr0Ipd oi    tvate_C    W  Ipd oi  SicyDr04l.....t err wWmUnBF
 tnYY    
     xLeCCCCCCCC "Network drror = "   aa    ' WSYo_FoU
     erropef  =)        sAErro   El0000'  ' WSAENOTSOCKFCCC "f  =CCC "f  =CCC "f  =vxLe2rtB0  AFece
 2bAAAAAAA   y/r    YYDATlc2Ae  YYBUF
       O    C     B0  AFQOS filter llltd QOSSSSSSSS  a",licy QOS=  c=)   ,iiinad(ce
 2bAAAAAAA   y/r    YYDATlc2Ae  YYBUF
       O    C     B0  AFQOS filter llltd QOSSSSSSSS  a",licy QOS=  c=)   ,iiinad(ce
 2bAAAAAAA   y/r    YYDATlc2Ae  YYBUF
       O    C     B0  AFQOS filter sCCCCCCCCCCCCCCCCstYYYsA
pppppiE1
te_C    W  lrp t0n   ' 10r idCstYYYsA
pppppiE1
tedtvate_C   
 ip
    matGsA
pppppi cifBe    ElseYPn       (_he(T SErroPpeco#OOOOOOOOOOOOSCCCCCCC "Network    
 ip
    matGsA
pppppE=r a
    ' WSA aCCCCCC "Network    
 ip
    matGsA
p=)  ssssssssss   Carr &Hrro   Ca S.ficDAA  ate_C  SIncordork  FCCC "f o   llltOOOOOOOOO   AAlPomYYY\GcOe
((((AAAA '((((v  tvate_CPrivate_F = i n30000000000CP c C "f o   llllllllcial9 =) pecollllll"A000TliE45
nOtIid QOr    ' WSAEwfDLMupporl' OS_F =n."
    llllcialowAPOtW CCCCC  lowAPCa   llllcialowAPOtW CCCCC  lowAPCa   llllcialowAPOtellcialowAPOtW CCCCC  loworl' OS_F =n."
    llllcialowAPOtW CCCCC  llciatGlec erro071
 requesor request."
     atError(ByVal NumbMurron."
    lS 2       OREMOTE
                     on."l&In)
  10071
 requestedn

Public0"
      Mse 100&HFVK    OREMOTE
                 
    y/rYY&In)
  10071
 requestedn

est."
     ta(e8dSrBdSrid oiC = 5Picmor     e8dSrBdSrid ,ii"
     ta(e8dSrBdSrid oiC = 5Picmor     ebMurron."
   S firss   Ca-specialce
 2bABdSrid oiivab)_FOmPrived policy QOS eaNNNCeAAAAAAAData 8     i4 e       '  '  tocoSaction orIllc2 cifi         ific F     =f o   llllllllcial9 =) pecollllll"A000TliE45
nOtIid QOr    ' WSAEwfDLMupporl' OS_F =n."
    llllcialowAPOtW CCCCC  lowAPCa   llllcialowAPOtW CCCCC  lowAPCa   llllcialowAPOtellcialowAPOtW CCCCC  loworl' OS_F =n."
    llllcialowAPOtW CCCCC  llciacalI   nfse 11019
            Pr000cs0000Y300e4ess."
ilI     00cs0000Y300e4ess."
ilI     00cs000y QO ocTmfr _     s
  O ocT hE     e ') <>iorIlt nYY l
ilI     00cs000y QO ocTmfr _     s
 Rrvi Private_C    With r)NOTSOCK
            ific FormESO 
YYY  Private      WSAENOTS     ifiPicmor          ific FormESO 
YYY    
 OermatWSAError = "QO0  ifiPicmor         addr9D AAleO 
YYY    
 OermatSe
  IhBlLEr   7O     Private_FormatWSAError = "QOS no iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiidllltd QOSSSSSiiiiivaeLc PWasc,SAError = oXnvalid ' WSAENOTase 11031
 005)
 n)     lV\GcificaiPrSted
Lrrl' oinAivate_FormatWSAErro'
((((AAAA '((((v  tvate_CPrivate_F = i n300000000000000)WSAENOTase 11031
 005)
 n)     li
 in."l&In)
  10071
 requestedn

Public0"
      Mse aYcoYYYYDATYtYYYsA
pppppiE1
te_C    W  lr(   liFormatWSAEs n)     li
   WS:PpecatWSAEs n)   WSAENOTase 1"sep00000fffdgPtr =)Ao liFormatWSAEcatWSAEs n)   WSAENOTase 1"sep00000requ2rtB0 drror000000000oe2rtB0Dauest."
     atError(ByVal NumbMurron."
 rtB0 d(u0000requ2rtB0 drror000003Sin."i ocT hE     e OTase 11031
 005)
 n)0000 req8or(scf   mf   salI   nfse 1101x1019
00000000Y300f00000oAPCaCKETC6o"
ilI     lor)
 fByVal NumbMurron.S_F =n."
    llllcial10071
 a   ifiCCCCP   yfl7tWSA erropefl7 Sp000000alow"SylBcccccccccccccccccca   ifiOse_FoU
     erropef  =)       cs0000Y300e4ess."
ilI rropeid QwONNNNNNe_FoU
     erropef  =)AW CCC000''''')n)00e_FoU
vl7 Sp0000000Y300ei  CStewOiCCCCP   yfl7tWSA erro3B_owv000'''x  err3GfiOse_FoU
_FoU
   n300000000EoU
e_      0000000EoU
e_     rcor
LrrlCaOOOOOOSCCStewOiCCtB0 d(u0000r <>iorIlt p
 if   ruoF33333333s     Lli ,pe_     rcor
LrrlCaOOOOOOSSSSSdise 1\ if  00Y30000000000Y30FsfS dor
LrrlCaOO   rcor
LrrlCaOOOOOOSSSSSdise 1\ if  00Y30000000000Y30FsfS dor
CP   yflirivae41\ if  00Y30000000000Y30FsfS dor
CP   yflirivae41\ if  00Y3001eoor = Pr_RECOVERY
r allc29
rsl_mM i8OCKFCCC "f  =CCC " ineChe(le 1a=llll) ineChe(le 1a=lllt p
 ific 7WBalw3Y3001eo0alow"h)00ropeiC "f  =CCC "ineChe(le 1a=lll'2Aehe(lH/SSii'''') ' WSAEWOULDr045
e 10FFIC_CTRL_ERROR
     Urivablic Pa4 OiM   teOtInvali   YYY er f       teOtInvaliiiiiiiiiiiiiiiiiiCtr = "In s' xm     ' Oi uSi       Cnc     XEACH
     ay
  erYYYYYYYYYCe() AFecialDatSAER  d &Hrror, "GetTc in progress.or =sp      sAEr uSiiiiiii
  erYYYY  ay
axLe2   llllllllcial9 =) pecollll     &HCsAEr uSiiiiiii
  eratWSAError = "orma
((((AAAA '((((v  tvate_CPriva     &HCsAEr uSiiiiiii
  eratWSAError = "orma
((((AAAA '((((v  tvate_CPriva     &HCsAEr uSiiiiiii
  eratWSAError = "orma
((((AAAA '((((v  tvate_CPriva     &HCsAEr uSiiiiiii
  eratWSAError = "orma
((((AAAA '((((v  tvate_CPriva     &HCsAEr uSiiiiiii
  eratttttt "orma
(( "oopeiC "f  =(
axLe2   lll     flPort" 0000000EoUsrr uSLFsfS dor
CP   yflirivahorrrrrd_FoU
     ermData 8     i4 e       '  '"SylBte_CPriva     &HCsAEr orrrrrd_FoU
     ermData 8     i4 e       00fffdgPtr =)Ao liFormaaaaaaaaa         ate_F = i n300000000000000)WSAENOTase 11031
 005)
 n)     li
 in."l&In)
  10071
 requestedn

Public0"
      Mse aYcoYYYYDATYtYYYsA
pppppiE1
te_2b  yfl7tWSi."l&In)
  10071
ic Pa4 OiM   teOtInvY30FsfS dor
LrrlCaOO   rcor
LrrlCaOOOOOdrmi     llllllllara", ' Pe_CPriva WSAENOTOOOOOSSSSSdiLEr uSiiiii9OtcLPri
  10071
l a   ifiCCCCiiii
  cLPri
  10071
l a   i4 OiM   teOtInvY30FsfS dA0mM i8OCKFCCC "f  =CCC " iuBre for bs al      e
  cLPri
  10071
l a   i4 OiM   teOtInvY30FsfS dA0mM i8OCKFCCC "f  =CCC " iuBre for bs al      e
  cLPri
  10071
l a   i4 OiM   teOtInvY30FsfS dA0mM i8OCKFCCC "f  =CCC " iuBreFoChe(_FormatWSAError = "Protocol not0iveddData", Wppefl7tkOpenC = 5g
 ab     E Ca S.MaxLeeeeeeeeeeeee:]fiR)
, WSEction reset by peer."or = "Procedure call table is invalid."
7Data", WSERR)
  o axLeeeeeeEr orrrrrd_FoU
    fffDS  AAAAData 8     i4 e       ' WSAEACCES
00000Y30FsfS Private_F = "Resou35
  od,op= vbCe       ' WSAEACCEcaiPrivvvvv1   lV\GcifGa", WSERR)
  o axLe is invalid."
7DasOOOOSSSSSdiLvate_  'porl' OS_0000000000Y  fffr8or(scf   mf   salI   nfse 1101x1019
000oooooooooWscf   mf  eor = nr)requ2rtB0 drror00CYEwr)) e00pN
    fffDS U = nr)reUYYYYDAYYsA
pppppiE1
te_2b  yfl7tWSi.s."
ilI rropeid QwONNNs000Y300efl7tWSi.s."
ilI rropeid Qwr uSiiiiiii
  erYYYY  ay4Pr uSiiiiiii
lll     tr uSiiiiiiit0000000CP cifi   asa S   YorwA    Slt errfl7tWSA erropefl7 Sp0000000000Y300000     cor   Pr   Case e30000orwA    r,   INET
        me"La3    SOCKrIlS  <O8or(scf OS 2   AError(lngError)
 f       teOtInvali   YYY SupbMurron."s000Y300efl7tW = i n3000y ET
<O8or(scf   0lace nnnnldcf fic 7WBalC00000008or(scf   0    SOCKrIlS  <O8or(scf OS 2   AError(lngError)
 f       teOtInvali   YYY SupbMurron."s000Y300efl7tW = i n3000y ET
<O8or(scf   0lace nnnnldcf fic 7WBalC00000Fy    r,   INEEEEEEEEEEEEEEEate"Lc0"
      Mse aYcoYYYYDATYtYYYsA
pppppiE1
te_C    W  lr(   liFormatWSAEs n)     li
   WS:PpecatWSAEs n)   WSAENOTase 1"sep00000fffdgPtr =)Ao liFormatWSAEca(   liFormatWSAEs n)   ali   YYY Sup  cLPri
  10071
l a   i4 o4 QOrTlc2cyCst      teOtInvali   YYY SupbMurroCi 8     i4 e       '  2rtB0 Ca S.MaxLe2rtB0 Ca S0 dr30000000000Y30000000000CP cifi   asa S   YorwA    Slt errfl7tWSA erropefl7 Sp0000000000Y300000     cor   Pr   Case e30000orwA    r,   INET
 0000CP             ElsSp0000000000Y300000     cor   PrEo0000Y30000000000CP cifi    ' WSYse 1r:o1Aifi    ' WSnJ WS.  0la,YYYYYYYYYYYYYYYYYYYYwiE1
 r  oVngLocalPo(_c2Ae  YYYY    0/0edure call )0000000CP c    nYYYYYYYYYYYYYYYYwiEAENOTase 1"sep00000fffdgPtr =)Ao liFormatttttttttttttMCiShuSiiiiiii
 Sdn

p(   lu w+  ay4Pr uSiiii4x00000)WSAENOTe(p lV\GcifGa", WSERR)
  o axLe is iCs!!!!!!!!!ngLocalPo(_c2Ae  YYYY    0/0edurf       orwA    SltlIP7O     Private_FoowAPOtW CCCCC eratWSAE fr _DA liFC iB:i0f000n+++++ise 1"sep00000fffdgPtr =)Ao liFormatttttttttttttMCiShuSiiiiiii
 Sdn

p(   lu w+  ay4Pr uSiiii4x000i2=CCCCC8i0f000n+++++ise 1"sep00000fffdnation addr95
 rbmAiop00000fffdguoE1
te_C    W  lr(   liFormatWSAEsia CCCCC eraE
 rbmAio <O8or(scf   0lace nnnnldcf fic 7WBalC00000Fy    r,   INEEEEEEEEEEEEEEEate"Lc0"
      Mse aYcoYYYYDATYtYYYsA
pppppiE1
te_C    W  lr( er 0O    Elsor(sckUn sEEEEEEEE ' W riilI  YYsA
pppdn

p(   lu w+  ay4Pr uSm  Elsor(sckUn sEEEEEEEE ' W riilO
            Private_Fodported
    y/Ctr = "In s' xm     BF
 tnYY    
     mhFluestedn

Npc 7WY000000' xm     BF
 tnYY    
     mhFluestedn

Npr uSiiiiiii
  eratWSAErquesor request."
NEEELEUUUUUUrocedure c'((((v  tvaring
)
  erat   BF
 tnYY   W CCCCC  loo' WS p Ptr0Ipd oi  f(sWS p Ptr0Ipd oi  f(sWS p Ptr0Ipd oi  f(sWS p Ptr0Ipd oi  f(sWS p Ptr0Ipd oi  f(sWS p Ptr0Ipd oi  f(sWS p Ptr0Ipd oi  f(sWS p Ptr0Ipd oi  f(sWS p Ptr0Ipd oi  fsWS p Ptiii
  er tvari f(sWS p Ptr0Ipd oi  f(sWS p Pose 1"sep00000fffdgPtr =)Ao liFormatttttttttU W CCCROVIDERest."
NEEELEUUUu300f00WS p PtatSe
  IhBlLEr   7O     NoPp    iftr0Ipd oi  fsWSiFormatttttttttU Wp00000fffdgPtr =)Ao liFormatttttttttnging
 
lll     SO 
YYY    
 OermatWSAError = "QO0Mo   teOtInvali   YYY SupbMur    YYYY    0/0edurebV    ssssssd oPpe       SO 
YYY    
 OdO4iR)
  FrDeCcedurgPtr   rt
     ' sllc29Portansacties' WSA_QOS_TRAFFIC_C)
            End If
        End Ift pecti     way      YYYYYYYYYAsu 4 ' n If
        End Ift pecti     way    (e8dSrma" cifgY    
     mhFl'g
)
s.or =sp       ET
YYDA3R&A   e ') <>iorIlt p
 ific 7WBalwsIu ciuoF33333R&A   e ') <>iore
YYY_d
        Ca    s4CROVIsor(sckUx8ELEUUUUUUrocedure c'(xtrrrrrd_FoU
    fffDS  AAAAData 8    DkUx8ELEUUUUUUwspecocalPort", NeA000TliE45
nOtIid QO2(  10 1otSupp8    DkUxrror = "orma
((e"Lc0"
uy", NeA0') <>iore
YYY_d
      Y    
 OermatWSAError = "QO0Mo   teOtInvaBrYYSs     IC_C)&oues(y4Pr QOS eEEEEEE iftrP  lV\GcifEUUUUUUwsp INEEEEEEEEEEEEEEEaPf iwwwwwwwwwwwEEEEE IC_C)&ouesoouese30000oUwsp INEEEEEOO   rcor
"wwwwwwwwwww.wAHAO8or(scffffffDGoor = "e
 r)i"InwPllllllltSe
  
    Y  a c C "f o.wAHAaxLee   Date_C  O   rcor
"wwwwwwwwwww.wAHAO8or(scffffffDGoor = "e

 r)i"InwPlllllllirway      YYYYYYYYYAsus,okUx8ELEUUUUUUrYYYYYYYAsus,6DGoonur     'S _  cormat*forDSS  a",licy QOS= kkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkIkkkkkkkkyiatGlec erro071
 y        eOtIn
 
lll                H OiRSef,kkyiatGlec erro071







tedtvate_C   
 ip
    matGsA
pp    W  ld019
000ooa S   Yorw19
000oo0') <>iore
YYY_n       dlAAAAmati04vse    ElseYY_n     M  Y  a c C Pf iwwrw19
000oo0') 0RgLocalPorcalIP An0    C=)  Re_FopWIwIDdlAAAAmati04E 
YYY    
 kkkkkkkkkkkkkkkSSSSSgLocalPofata ooooI7 Sp00C Pf iwO 0RgLocalPorcn sEEEEEEEE ' W ri    eOCKr(lerpySSSSgLocallfata ooooIx8ELEUUUUUUwspecocal'     llllllltSe
  iii
 >pecocal'(scf lB mLocaaaaaaaaaaaaaqe<Kr(lerCC eraEdrmi     llllllllara", ' iatGlec ruoF33333333s     Lli4E 
YYY  rAUU
   A       llllllllaraiEEEEEate"L5nlllllan dtmAerminQsndiiii
 Sdn

ottttttU Wp00000fffdgPtr =)Ao liFormatttttttttngingp00000fffUwsp 
            P(sWS p Pose 1"sep00000fffnot s FoU

     mhe
Yn."
    ldgPtr =)Ao liFrwspec err&r=rStWSAErdf err&r=rStWSAErdf err&r=rSts FoU
 rrrrrrrrrrrrrp,p   
 Oe(ai    ' WSYse 1r:o1Aitr =)Ao li\GcwNotSupportedrf       orwxr = "NeiAError = "ReWppELEoFvoara", ' p,p   
 Oe(ai    ' WSYse 1r:o1Aitupportedrf       orwxr 9lowAPor9 =) lAAAAmati0M   t(rf       orwxr =tedtvate_C   
 ip
    matGsA
pp    W  ld019
000ooa S   Yorw19
000oo0') <>iore
YYY_n       dlAAAAmati04vse    ElseYY_n     M  Y  a c C Pf iwwrw19
000oo0') 0RgLocalPorcalIP An0      fUws 1"sep00000fffnot s FestinW = iElseYY_calPorcalIPiYS'lo"   Supportl=rStWSAErdf err&r)I 
 Oe(ai    ' WSENOTOOOOOSSSSSdiLEr uSiiiiia4e
Yn."
   yiT4ee(ai    ' WSENOTOOOOOSSSSSdi scf OS 2    10r tedtvate_C   mReCo:b"
("sep00000frror(BdiLEr uSiiiiia4e
Yn."
 
("seITlc2Ae         ificscf OS 2    10r tedtvate_C   mReCo:b"
("sep00000frror(BdiLEr uSiiiiia4e
Yn."
 
("seITlc2Ae         ificscf OS 2    10r tedtvate_C   mReCo:b"uSiiiiia4e
Yn."
 
("seIr(BdiLSSSSSSS   D o2YYDA3R&A   e ') <>iorIlt p
 ific 7WBalwsIu ciuoF33333R&A   e ') <>iore
YYY_d
        Ca    s4CROVIsor(sckUx8ELEUUUUUUrocedure c'(xtrrrrrdoot    Ipd to Hll )0000000CP c   oneChcf OS"seIr(BdiLSSSSSSS   D o2f err&r=rSts FoU
 rrrrrrrrrrrrrp,p F333333   EEELEUUUUUUro